/*---------------------------------------------------------------------------*\

	FILE....: openpri.cpp
	AUTHOR..: Ben Kramer
	DATE....: 29/07/2004
	DESC....: Main part of the OpenPri driver

\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

         Voicetronix Voice Processing Board (VPB) Software

         Copyright (C) 1999-2004 Voicetronix www.voicetronix.com.au

         This library is free software; you can redistribute it and/or
         modify it under the terms of the GNU Lesser General Public
         License as published by the Free Software Foundation; either
         version 2.1 of the License, or (at your option) any later version.

         This library is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty of
         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
         License along with this library; if not, write to the Free Software
         Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	 USA

\*---------------------------------------------------------------------------*/

#include "utilfuncs.h"
#include "sangoma.h"
#include "openpri.h"
#include "hostvox.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <limits>


extern "C" {
#include "alawmulaw.h"
#include "dtmf.h"
#include "toneg.h"
#include "toned.h"
#include "echo.h"
#include "libpri/pri_q931.h"
#include "libpri/pri_q921.h"
}

#include "mess.h"
#include "messages.h"
#include "generic.h"
#include "dspfifo.h"
#include "mapdev.h"
#include "comm.h"
#include "wobbly.h"


CONF_INFO * config_p;

typedef struct {
	int card;
	int channel;
} CON_THREAD_INFO;

struct cref_info {
	int cref;
	struct isdncid cid;
	int layer1;             // Encoding as from Call setup or for next call
	int transcap;           // Transmit capabilities form Call setup or for next call
	void *next;
};



/* VPB Driver stuff */
#define ALAW_CODES	256
#define MAX_BOARDS	12
#define FRAME 		8
#define PRI_PORTS	32
//#define NBUF		160
//#define NBUF		80
//#define NBUF		40
//#define NBUF		16
#define NBUF		8
#define HOSTDSP_SLEEP	5


extern int wanconnect(char * card, char * port);

static void priOpenPorts(VPBREG * v);
static void prepare_rec_buf(unsigned char alawbuf[NBUF],  VPBREG * v, int ch, int encoding);
static void prepare_rec_buf_log(unsigned char alawbuf[NBUF],unsigned char alawbuf2[NBUF], VPBREG * v, int ch, int encoding);
static void prepare_play_buf(unsigned char alawbuf[NBUF],  VPBREG * v, int ch, int encoding);
static void toned_callback(word mess[],void *board);
static void priSimulateDSP(void *d);
static void priProcessCommands(VPBREG * v);
static void priClosePorts(VPBREG * v);
static void priRunDchannel(void *d);
static void priRunDchannelLog(void *d);
static void handle_pri_event(struct pri *pri, pri_event *e, VPBREG * v);
static void handle_pri_event_log(struct pri *pri, pri_event *e, void *d);
static int hw_gain_vector(float g, short *v, int n);
static int make_isdn_call(word * d, VPBREG * v);
static int findfreechan(VPBREG *v);

static void *audio_thread_multi(void *pv);
static void *audio_thread_log_multi(void *pv);
pthread_t audio_thread_t[32];

/* Echo stuff */
#define ERL_SIM         0.1     // default 1, to emulate a hybrid turn on loopback and set to 0.1
#define LUTOS   32768		// Look Up Table Off Set


/* libpri stuff */
#define MAX_CHAN		32
#define	DCHANNEL_TIMESLOT	16

/* Channel state defines */

#define CH_IDLE		0
#define CH_RING		1
#define CH_PROCEEDING	2
#define CH_RINGING	3
#define CH_SETUP	4
#define CH_NEEDANS	5
#define CH_NEEDHANG	6
#define CH_HUNGEDUP	7
#define CH_NEEDRING	8
#define CH_TERMINATING	9
#define CH_CONNECTED	10
#define CH_NEEDTERMINATING	11

/*--------------------------------------------------------------------------*\

	FUNCTION.: priHostDSPOpen
	AUTHOR...: Ben Kramer
	DATE.....: 16/11/04


\*--------------------------------------------------------------------------*/

void OpenPri::priHostDSPOpen()
{ //{{{
	int   i,x;
//	numbd = anumbd;
	int islog=0;
	mprintf("[%d/-] Initialising PRI Host DSP ...\n", vr->cardnum);
	finish = 0;
	audio_multi=0;

	OP_CARD *cardinfo = (OP_CARD *)vr->cardinfo;
	int dchan = cardinfo->d_channel;
	int isdn_num=0;
	int iface_num=0;

	int numch = vr->numch;
        vr->toneg = new void*[numch];
        vr->toned = new void*[numch];
	pri_chan * chans = new pri_chan[numch];
	vr->chans = chans;
	vr->hostdsp = NULL;

        if (cardinfo->node == ISDN_NODE_LOG) {
                islog=1;
		mprintf("[%d/-] This card is configured for logging\n", vr->cardnum);
        }

	mprintf("[%d/-] This card has [%d] channels\n", vr->cardnum,vr->numch);
	for(i=0 ; i<numch; ++i) {
		++iface_num;
//		mprintf("[%d/%d] Starting channel indexed %d ...\n",vr->cardnum,isdn_num,i);
		toneg_open(&vr->toneg[i], i );
		toned_open(&vr->toned[i], i , &toned_callback, (void *)vr);

		/* setup chans structure */
		chans[i].state = CH_IDLE;
		chans[i].run = 0;
		// layer1 ?  = encoding ??
		chans[i].dtmf_dec = dtmf_init();
		chans[i].hand_num = -1;	// set the handle number so that we know who we are
		chans[i].chan_num = isdn_num;	// Channel number (not true ISDN as doesnt map above d-chan)
		chans[i].cref = 0;
		chans[i].vox_states = NULL;
		#ifdef CASLOG
		if(islog){
			hostvox_open(&chans[i].vox_states,i);
		}
		#endif
		if (iface_num >= dchan){
			chans[i].iface = &cardinfo->iface[iface_num+1];
		}
		else {
			chans[i].iface = &cardinfo->iface[iface_num];
		}
		chans[i].encoding = chans[i].iface->encoding;
		chans[i].echo_enable = 0;
		chans[i].cause = 16;	// Set it as normal clear down
		chans[i].cid = NULL;
		chans[i].v = vr;
		chans[i].call = NULL;
		chans[i].tx_fifo = NULL;
		if(echo_open(&chans[i].echo_states)){
			mprintf("[%d/%d] Unable to set up echo cancellor on channel\n",
				vr->cardnum,chans[i].chan_num);
			chans[i].echo_enable=-1;
		}
		for(x=0; x < NSAM + ECHO_TAPS; ++x) {
			chans[i].echoy[x] = 1.0;
		}
		for(x=0; x < NSAM; ++x) {
			chans[i].echos[x] = 1.0;
			chans[i].echoe[x] = 1.0;
		}
//		mprintf("Started channel [%d] on board [%d] with handle [%d] indexed [%d]\n",
//			chans[i].chan_num, vr->cardnum, chans[i].hand_num, i);
//		mprintf("[%d/%d] Started channel with handle [%d] indexed [%d]\n",
//			vr->cardnum, chans[i].chan_num, chans[i].hand_num, i);
		mprintf("[%d/%d] Started ISDN channel ...\n",vr->cardnum,isdn_num);
		++isdn_num;
	}

	// Check if we are running audio_thread_multi
	// run through all channels and compare the interfaces,
	// counting the number of audio_thread_multi required
	audio_multi=1;
	for(i=0 ; i<numch-1; ++i) {
		if (strcmp((char*)(chans[i].iface->name),(char*)(chans[i+1].iface->name))){
			mprintf("Audio Multi [%d] [%s] != [%s]\n",audio_multi,
				chans[i].iface->name,chans[i+1].iface->name);
			++audio_multi;
		}
	}
	audio_thread_multi_info = new AUDIO_THREAD_MULTI[audio_multi];
	x=0;
	audio_thread_multi_info[x].run=1;
	audio_thread_multi_info[x].start_channel=0;
	audio_thread_multi_info[x].end_channel=0;
	audio_thread_multi_info[x].v=vr;
	audio_thread_multi_info[x].card = chans[0].iface->card;
	audio_thread_multi_info[x].interface = chans[0].iface->name;
	for(i=0 ; ((i<numch-1) && x<audio_multi); i++) {
		if (strcmp((char*)(chans[i].iface->name),(char*)(chans[i+1].iface->name))){
			audio_thread_multi_info[x].end_channel=i;
			// Start up the audio thread
			pthread_create(&audio_thread_t[x], NULL,
				       islog ? audio_thread_log_multi : audio_thread_multi,
				       (void*)&audio_thread_multi_info[x]);
			mprintf("[%d/-] Started Audio Thread Multi [%d]...\n",vr->cardnum,x);

			// prepare for next audio thread if there is one
			x++;
			if (x<audio_multi){
				audio_thread_multi_info[x].run=1;
				audio_thread_multi_info[x].start_channel=i+1;
				audio_thread_multi_info[x].end_channel=i+1;
				audio_thread_multi_info[x].v=vr;
				audio_thread_multi_info[x].card = chans[i+1].iface->card;
				audio_thread_multi_info[x].interface = chans[i+1].iface->name;
			}
		}
	}
	// Start up the last audio thread
	audio_thread_multi_info[x].end_channel=i;
	pthread_create(&audio_thread_t[x], NULL,
		       islog ? audio_thread_log_multi : audio_thread_multi,
		       (void*)&audio_thread_multi_info[x]);
	mprintf("[%d/-] Started Audio Thread Multi [%d]...\n",vr->cardnum,x);

	priOpenPorts(vr);

	GenericInitializeCriticalSection(&mutex);
	Generic_beginthread(priSimulateDSP, 0, this);
	if(islog){
		Generic_beginthread(priRunDchannelLog, 0, this);
	} else {
		Generic_beginthread(priRunDchannel, 0, this);
	}
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priHostDSPClose
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04

	Constructor to stop host DSP processing.

\*--------------------------------------------------------------------------*/

void OpenPri::priHostDSPClose()
{ //{{{
	pri_chan *chans = (pri_chan *)vr->chans;
	int cardnum = vr->cardnum;
	int i = 0;

	mprintf("[%d/-] Closing Pri Host DSP...\n",cardnum);
	finish = 1;
	for(; i < audio_multi; ++i) audio_thread_multi_info[i].run = 0;
	i = 10;
	while( active_threads && --i ) GenericSleep(500);
	if( i == 0 )
		mprintf("[%d/-] FAILED to stop all threads, %d still active\n",
			cardnum, active_threads);
	//else  mprintf("[%d/-] stopped all threads after %d attempts\n",
	//	        cardnum, 100 - i);

	mprintf("[%d/-] Closing down ports...\n",cardnum);
	priClosePorts(vr);
	mprintf("[%d/-] Releasing per channel data structures...\n",cardnum);
	for(i=0; i < vr->numch; ++i) {
		dtmf_close(chans[i].dtmf_dec);
		toneg_close(vr->toneg[i]);
		toned_close(vr->toned[i]);

		chans[i].state = CH_IDLE;
		//mprintf("[%d/%d] c = %p\n", cardnum, i, chans[i].call);
		pri_destroycall(pri, chans[i].call);
		q931_destroycall(pri2, chans[i].cref);
		chans[i].call = NULL;

		echo_close(chans[i].echo_states);
		chans[i].echo_enable = -1;
		delete chans[i].cid;
		chans[i].cid = NULL;
	}
	mprintf("[%d/-] Released channel data structures...\n",cardnum);
	mprintf("[%d/-] Freeing OpenPri structures...\n",cardnum);

	if( pri )  q921_discard_retransmissions(pri);
	if( pri2 ) q921_discard_retransmissions(pri2);

	delete [] vr->toneg;                vr->toneg = NULL;
	delete [] vr->toned;                vr->toned = NULL;
	delete [] chans;                    chans = NULL;
	delete [] audio_thread_multi_info;  audio_thread_multi_info = NULL;

	GenericDeleteCriticalSection(&mutex);
	mprintf("[%d/-] Closed Pri Host DSP!\n",cardnum);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priOpenPorts
	AUTHOR...: Ben Kramer
	DATE.....: 16/11/04

	Initialises PRI ports.

\*--------------------------------------------------------------------------*/

void priOpenPorts(VPBREG * v)
{ //{{{
	int     i;
	int cardnum = v->cardnum;

	for(i=0; i<v->numch; i++) {
		// init DSP (audio) FIFOs
		v->rxdf[i] = new DspFifo(DSPFIFO_UP, v->szrxdf[i],0);
//		mprintf("record side writing [%02d] 0x%x\n", i, (int)v->rxdf[i]);
		v->txdf[i] = new DspFifo(DSPFIFO_DOWN, v->sztxdf[i],0);
//		mprintf("play side reading [%02d] 0x%x\n", ch, (int)v->txdf[i]);
		mprintf("[%d/%d] DSP Audio FIFOs opened OK\n",cardnum,i);
	}
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priClosePorts
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04

	Shuts down PRI ports.

\*--------------------------------------------------------------------------*/

void priClosePorts(VPBREG * v)
{ //{{{
	int     i=0;
	int cardnum = v->cardnum;

	for(; i < v->numch; ++i) {
		delete v->rxdf[i]; v->rxdf[i] = NULL;
		delete v->txdf[i]; v->txdf[i] = NULL;
		mprintf("[%d/%d] DSP Audio FIFOs closed OK\n",cardnum,i);
	}
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: prepare_play_buf
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04

	Reads bufs of play samples, and formats them ready for playing.
	Also iterates tone generation.

\*--------------------------------------------------------------------------*/

void prepare_play_buf(unsigned char alawbuf[NBUF],  VPBREG * v, int ch, int encoding)
{ //{{{
	int	ret,retread;
	short	linbuf[NBUF];
	word	mess[COMM_MAX_MESSDSPPC];
	int	i,chdev;
	pri_chan *chans = (pri_chan *)v->chans;
	//OpenPri* pThis = (OpenPri *)v->hostdsp;
	//void     **toneg = pThis ? pThis->toneg : NULL;
	int idx;

	try {
		idx = get_idx_from_ch(v,ch);
	} catch ( const Wobbly &w ) {
		mprintf("prepare_play_bug: Cant work out IDX from ch[%d]!\n",ch);
		return;
	}

	// Read play fifos, add in tone, convert to alaw
	memset(linbuf, 0, sizeof(linbuf));
	retread = v->txdf[ch]->DSPRead( (word*)linbuf, NBUF);
	if(VPB_RAW == encoding){
		for(i=0;i<NBUF;i++){
			alawbuf[i]=linbuf[i];
		}
		return;
	}
/*
	if (retread != 0){
		mprintf("prepare_play_buf: FIFO empty on channel[%d]!\n",ch);
	}
*/
//	mprintf("Playing tone....");
	//chdev = mapdevtohndle(v->cardnum,ch);
	//ret = toneg_play(toneg[chdev], linbuf, NBUF);
	//ret = toneg_play(toneg[idx], linbuf, NBUF);
	ret = toneg_play(v->toneg[idx], linbuf, NBUF);
	if (ret == -1) {
		mprintf("Finished playing tone on ch[%d] idx[%d] chdev[%d] card[%d]...\n",
			ch,idx,chdev,v->cardnum);
		mess[0] = DSP_LTONEG;
		mess[1] = DSP_TONEG;
		mess[2] = ch;
		v->upmess->Write(mess, mess[0]);
	}
//	mprintf("done\n");
	/* Do echo stuff */
	if (chans[idx].echo_enable > 0){
		/* copy frame to echo buffer */
		for (i=0;i<NBUF;i++)
			chans[idx].echoy[ECHO_TAPS+i]=linbuf[i];
	}

	/* Apply the Hardware gain */
	hw_gain_vector(v->defPlayGain,linbuf, NBUF);
	/* Needs to check if we are alaw or mu-law */
	if (encoding == VPB_ALAW){
		alaw_encode((char *)&alawbuf[0], linbuf, NBUF);
	}
	else if (encoding == VPB_MULAW){
		mulaw_encode((char *)&alawbuf[0], linbuf, NBUF);
	}
	else {
		mprintf("[%d/%d] Encoding format[%d] not supported!\n",
			v->cardnum,chans[idx].chan_num,encoding);
	}
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: prepare_rec_buf
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04

	Transfers buffers of recorded samples to the driver FIFOs.  Also
	runs DTMF decoders on each channel.

\*--------------------------------------------------------------------------*/

void prepare_rec_buf(unsigned char alawbuf[NBUF], VPBREG * v, int ch, int encoding)
{ //{{{
	int	digits_ret;
	short	linbuf[NBUF];
	char	key[2];
        word	mess[COMM_MAX_MESSDSPPC];
	int	i,ret;
	int	keydown=0;
	//OpenPri *pThis = (OpenPri *)v->hostdsp;
	pri_chan *chans = (pri_chan *)v->chans;
	int idx;

	try {
		idx = get_idx_from_ch(v,ch);
	} catch ( const Wobbly &w ) {
		mprintf("prepare_rec_bug: Cant work out IDX from ch[%d]!\n",ch);
		return;
	}

	if(VPB_RAW == encoding){
		/* Just write it to the buffer */
		for(i=0;i<NBUF;i++){
			linbuf[i]=alawbuf[i];
		}
		v->rxdf[ch]->DSPWrite( (word*)linbuf, NBUF);
		if (chans[idx].tx_fifo != NULL){
			chans[idx].tx_fifo->DSPWrite( (word*)linbuf, NBUF);
		}
		return;
	}
	// convert to linear and send to fifos
	/* Needs to check if we are alaw or mu-law !! */
	if (encoding == VPB_ALAW){
		alaw_decode(linbuf, (char *)&alawbuf[0], NBUF);
	}
	else if (encoding == VPB_MULAW){
		mulaw_decode(linbuf, (char *)&alawbuf[0], NBUF);
	}
	else {
		mprintf("[%d/%d] Encoding format[%d] not supported!\n",
			v->cardnum,chans[idx].chan_num,encoding);
	}

	/* Apply the Hardware gain */
	hw_gain_vector(v->defRecordGain,linbuf, NBUF);

	/* Do echo stuff */
	if (chans[idx].echo_enable > 0){
		/* copy frame to echo buffer */
		for (i=0;i<NBUF;i++)
			chans[idx].echos[i]=linbuf[i];
		/* do echo canceling */
		echo(chans[idx].echo_states,chans[idx].echoe,&(chans[idx].echoy[ECHO_TAPS]),chans[idx].echos,NBUF);
		/* copy echo buf back */
		for (i=0;i<NBUF;i++)
			linbuf[i]=LUTOS + (short)chans[idx].echoe[i];
	}

	/* for bridging and loop back  */
	if (chans[idx].tx_fifo != NULL){
		chans[idx].tx_fifo->DSPWrite( (word*)linbuf, NBUF);
	}
	/* write it to the buffer */
	v->rxdf[ch]->DSPWrite( (word*)linbuf, NBUF);

	// DTMF processing
	digits_ret = dtmf_decoder(chans[idx].dtmf_dec, key, 2, linbuf, NBUF, &keydown);
	if (digits_ret) {
		// it should be impossible to detect > 1 digit in NBUF sample buffer
		//assert(digits_ret == 1);
//		mprintf("ch = %d bd = %d DTMF[%d].....\n", chans[idx].chan_num, v->cardnum,key[0]);
		mess[0] = DSP_LDTMF;
		mess[1] = DSP_DTMF;
		mess[2] = chans[idx].chan_num;
		mess[3] = key[0];
		ret = v->upmess->Write(mess, mess[0]);
	}
	else if (keydown){
		mess[0] = DSP_LDTMF_DOWN;
		mess[1] = DSP_DTMF_DOWN;
		mess[2] = chans[idx].chan_num;
		mess[3] = keydown;
		ret = v->upmess->Write(mess, mess[0]);
	}

	// tone decoder processing
	toned_analyse(v->toned[idx], linbuf, NBUF);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: prepare_rec_buf_log
	AUTHOR...: Ben Kramer
	DATE.....: 20/07/05

	Transfers buffers of recorded samples to the driver FIFOs.  Also
	runs DTMF decoders on each channel. Takes the two audio streams from
	the audio log thread.

\*--------------------------------------------------------------------------*/

void prepare_rec_buf_log(unsigned char alawbuf[NBUF],
			 unsigned char alawbuf2[NBUF],
			 VPBREG *v, int ch, int encoding)
{ //{{{
	int	digits_ret;
	short	linbuf[NBUF];
	short	linbuf2[NBUF];
	int	sumbuf;
	char	key[2];
        word	mess[COMM_MAX_MESSDSPPC];
	int	i,ret;
	int	keydown=0;
	//OpenPri *pThis = (OpenPri *)v->hostdsp;
	pri_chan *chans = (pri_chan *)v->chans;
	int idx;

	try {
		idx = get_idx_from_ch(v,ch);
	} catch ( const Wobbly &w ) {
		mprintf("prepare_rec_bug_log: Cant work out IDX from ch[%d]!\n",ch);
		return;
	}

	if(VPB_RAW == encoding){
		/* Just write it to the buffer */
		for(i=0;i<NBUF;i++){
			linbuf[i]=alawbuf[i] + alawbuf2[i];
		}
		v->rxdf[ch]->DSPWrite( (word*)linbuf, NBUF);
		if (chans[idx].tx_fifo != NULL){
			chans[idx].tx_fifo->DSPWrite( (word*)linbuf, NBUF);
		}
		return;
	}
	// convert to linear and send to fifos
	/* Needs to check if we are alaw or mu-law !! */
	if (encoding == VPB_ALAW){
		alaw_decode(linbuf, (char *)&alawbuf[0], NBUF);
		alaw_decode(linbuf2, (char *)&alawbuf2[0], NBUF);
	}
	else if (encoding == VPB_MULAW){
		mulaw_decode(linbuf, (char *)&alawbuf[0], NBUF);
		mulaw_decode(linbuf2, (char *)&alawbuf2[0], NBUF);
	}
	else {
		mprintf("[%d/%d] Encoding format[%d] not supported!\n",
			v->cardnum,chans[idx].chan_num,encoding);
	}
	/* Converge audio streams */
	for(i=0;i<NBUF;i++){
		sumbuf=linbuf[i] + linbuf2[i];
		if (sumbuf > 32767){
			sumbuf=32767;
		} else if (sumbuf < -32768){
			sumbuf=-32768;
		}
		linbuf[i]=sumbuf;
	}

	/* Apply the Hardware gain */
	hw_gain_vector(v->defRecordGain,linbuf, NBUF);

	/* write it to the buffer */
	v->rxdf[ch]->DSPWrite( (word*)linbuf, NBUF);

	// DTMF processing
	digits_ret = dtmf_decoder(chans[idx].dtmf_dec, key, 2, linbuf, NBUF, &keydown);
	if (digits_ret) {
		// it should be impossible to detect > 1 digit in NBUF sample buffer
		//assert(digits_ret == 1);
//		mprintf("ch = %d bd = %d DTMF[%d].....\n", chans[idx].chan_num, v->cardnum,key[0]);
		mess[0] = DSP_LDTMF;
		mess[1] = DSP_DTMF;
		mess[2] = chans[idx].chan_num;
		mess[3] = key[0];
		ret = v->upmess->Write(mess, mess[0]);
	}
	else if (keydown){
		mess[0] = DSP_LDTMF_DOWN;
		mess[1] = DSP_DTMF_DOWN;
		mess[2] = chans[idx].chan_num;
		mess[3] = keydown;
		ret = v->upmess->Write(mess, mess[0]);
	}
	if (chans[idx].vox_states){
		ret = hostvox_analyse(chans[idx].vox_states,linbuf,NBUF);
		if (ret == 1){ // Vox OFF
			mess[0] = DSP_LVOXOFF;
			mess[1] = DSP_VOXOFF;
			mess[2] = chans[idx].chan_num;
			ret = v->upmess->Write(mess, mess[0]);
			mprintf("ch = %d bd = %d VOX OFF\n", ch, v->cardnum);
		}
		else if (ret ==2){ // Vox ON
			mess[0] = DSP_LVOXON;
			mess[1] = DSP_VOXON;
			mess[2] = chans[idx].chan_num;
			ret = v->upmess->Write(mess, mess[0]);
			mprintf("ch = %d bd = %d VOX ON\n", ch, v->cardnum);
		}
	}

	// tone decoder processing
	toned_analyse(v->toned[idx], linbuf, NBUF);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: toned_callback
	AUTHOR...: David Rowe
	DATE.....: 14/1/03

	This function is called by the tone detector when a tone is detected
	or a tone detector log message is generated.

\*--------------------------------------------------------------------------*/

void toned_callback(word mess[],void *board)
{ //{{{
	VPBREG *vr = (VPBREG *)board;

	// convert from  idx number to ch

//	mprintf("Got Tone[%d] from[%d]\n",mess[3],mess[2]);
	int idx = mess[2];
	pri_chan *chans;
	chans = (pri_chan *)vr->chans;

	mess[2] = chans[idx].chan_num;
	vr->upmess->Write(mess, mess[0]);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priSimulateDSP
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04

	Simulates the DSP on host processing cards:
	- polls card for signalling events
	- audio signal I/O is handled by different threads

\*--------------------------------------------------------------------------*/

void priSimulateDSP(void *d)
{ //{{{
	OpenPri* pThis = (OpenPri *)d;
//	VPBREG *v = pThis->vr;
//	GENERIC_CRITICAL_SECTION mutex = pThis->mutex;
//	int	*active_threads = &pThis->active_threads;
//	int 	*finish = &pThis->finish;

	mprintf("[%d/-] Starting PRI SimulateDSP\n",pThis->vr->cardnum);
	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	while(!pThis->finish) {
		priProcessCommands(pThis->vr);
		GenericSleep(HOSTDSP_SLEEP);
	}

	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Clean up thread
	pthread_detach(pthread_self());

	mprintf("[%d/-] Finishing PRI SimulateDSP\n",pThis->vr->cardnum);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priProcessCommands
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04

	Reads commands sent to "DSP" and processes them.

\*--------------------------------------------------------------------------*/

void priProcessCommands(VPBREG * v)
{ //{{{
	int	read;
//	int	config, lsf_ms, id; 
	int	ch, idx;
	int 	islog=0;
	word	mess[COMM_MAX_MESSPCDSP];
	word	rmess[COMM_MAX_MESSDSPPC];
	isdncid	*tmpp;
	int	ret;

	pri_chan *chans = (pri_chan *)v->chans;
	OP_CARD *cardinfo = (OP_CARD *)v->cardinfo;
	int cardnum = v->cardnum;
	OpenPri* pThis = (OpenPri *)v->hostdsp;
	void     **toneg;
	struct pri *pri;

	if (pThis != NULL){
		toneg = pThis->toneg;
		pri = pThis->pri;
	}
	if (cardinfo->node == ISDN_NODE_LOG) islog=1;

	read = v->dnmess->Read(&mess[0], 1);
	if (read != OK) {
		if (read != FIFO_EMPTY)
			mprintf("[%d/-]Error reading PRI message queue!\n",cardnum);
		return;
	}
	read = v->dnmess->Read(&mess[1], mess[0]-1);
	if( read != OK ){
		mprintf("[%d/-] Error on second read of fifo!\n",cardnum);
		return;
	}
	if( mess[0] < 3 ) {
		mprintf("[%d/-] Ignoring short fifo message %d < 3 [%#x]\n",
			cardnum, mess[0], mess[1]);
		return;
	}
	ch = mess[2];

	if( mess[1] != PC_TONEG ){
	    try {
		idx = get_idx_from_ch(v, ch);
	    } catch ( const Wobbly &w ) {
		mprintf("priProcessCommands: Cant work out IDX from ch[%d][%d][%d]!\n",
			mess[0],mess[1],mess[2]);
		return;
	    }
	}
	//mprintf("message bytes[%d]: on ch[%d]\n",mess[0],mess[2]);
	if (islog){
	//	mprintf("[%d/-] Ignoring command, in logging mode \n",cardnum);
		if( mess[1] == PC_VOX_UPDATE ) {
			assert(mess[0] == PC_LVOX_UPDATE);
			if(chans[idx].vox_states != NULL){
				hostvox_update_levels(chans[idx].vox_states, mess);
				//mprintf("[%d/-]Updating vox information\n",cardnum);
			} else  mprintf("[%d/-]No vox_states !\n",cardnum);
		}
		return;
	}

	switch(mess[1]) {
	    case PC_CODEC_ONHK:
		assert(mess[0] == PC_LCODEC_ONHK);
		if (chans[idx].state != CH_IDLE){
			mprintf("[%d/%d] Hanging up call\n",cardnum,chans[idx].chan_num);
			/* Stop Audio Thread */
			chans[idx].run=0;
			/* send Pri Message */
			ret = pri_hangup(pri,chans[idx].call,PRI_CAUSE_NORMAL_CLEARING);
			if (ret != 0){
				mprintf("[%d/%d] Error hanging up call\n",
					cardnum,chans[idx].chan_num);
			}
			/* Clear out caller ID information */
			delete chans[idx].cid;
			chans[idx].cid = NULL;

			pri_destroycall(pri,chans[idx].call);
			chans[idx].call = NULL;
			chans[idx].state = CH_IDLE;
		}
		//else mprintf("[%d/%d] Hanging up call - nothing to do, IDLE\n",
		//	       cardnum,chans[idx].chan_num);
		break;

	    case PC_CODEC_OFFHK:
		assert(mess[0] == PC_LCODEC_OFFHK);
		ret = -1;
		if (chans[idx].state == CH_RING){
			mprintf("[%d/%d] Answering call \n",cardnum,chans[idx].chan_num);
			/* send pri message */
			ret = pri_answer(pri,chans[idx].call,chans[idx].iface->channel,0);
			if (ret == 0) {
				chans[idx].state = CH_CONNECTED;
				/* start audio thread */
				chans[idx].run=1;
				//pthread_create(&audio_thread_t[idx], NULL,
				//		 audio_thread,(void*)&chans[idx]);
			} else {
				mprintf("[%d/%d] Failed to  answer call!\n",
					cardnum,chans[idx].chan_num);
				rmess[0] = DSP_LISDN_ANS_FAIL;
				rmess[1] = DSP_ISDN_ANS_FAIL;
				rmess[2] = chans[idx].chan_num;
				rmess[3] = ret;
				v->upmess->Write(rmess, rmess[0]);
			}
		} else {
			mprintf("[%d/%d]Cant answer call",cardnum,chans[idx].chan_num);
			mprintf(" not in ring state (state[%d])\n",chans[idx].state);
		}
		break;

	    case PC_SPI_BRIDGE:
		assert(mess[0] == PC_LSPI_BRIDGE);
		chans[idx].tx_fifo = v->txdf[mess[3]];
		try {
			idx=get_idx_from_ch(v,mess[3]);
		} catch ( const Wobbly &w ) {
			mprintf("priProcessCommands:PC_SPI_BRIDGE Cant work out IDX from ch[%d]!\n",
				mess[3]);
			return;
		}
		chans[idx].tx_fifo = v->txdf[ch];
		break;

	    case PC_SPI_BRIDGE_OFF:
		assert(mess[0] == PC_LSPI_BRIDGE_OFF);
		chans[idx].tx_fifo = NULL;
		try {
			idx=get_idx_from_ch(v,mess[3]);
		} catch ( const Wobbly &w ) {
			mprintf("priProcessCommands:PC_SPI_BRIDG_OFF Cant work out IDX from ch[%d]!\n",
				mess[3]);
			return;
		}
		chans[idx].tx_fifo = NULL;
		break;

	    case PC_SPI_SBRIDGE:
		assert(mess[0] == PC_LSPI_SBRIDGE);
		chans[idx].tx_fifo = v->txdf[mess[3]];
		break;

	    case PC_SPI_SBRIDGE_OFF:
		assert(mess[0] == PC_LSPI_SBRIDGE_OFF);
		chans[idx].tx_fifo = NULL;
		break;

/*
	    case PC_CONF_JOIN:
		assert(mess[0] == PC_LCONF_JOIN);
		hdaports[offset + mess[2]].conf_join(mess[3]);
		break;

	    case PC_CONF_LEAVE:
		assert(mess[0] == PC_LCONF_LEAVE);
		hdaports[offset + mess[2]].conf_leave();
		break;
*/

	    case PC_CODEC_RECGAIN:
		assert(mess[0] == PC_LCODEC_RECGAIN);
		break;

	    case PC_CODEC_PLAYGAIN:
		assert(mess[0] == PC_LCODEC_PLAYGAIN);
		break;

	    case PC_TONEG:
		assert(mess[0] == PC_LTONEG);
		pThis = (OpenPri *)v->hostdsp;
		toneg = pThis->toneg;
		try {
			idx = get_idx_from_ch(v,mess[12]);
		} catch ( const Wobbly &w ) {
			mprintf("priProcessCommands:PC_TONEG Cant work out IDX from ch[%d]!\n",
				mess[12]);
			return;
		}
		mprintf("[%d/%d] Starting Tone generation idx(%d)\n",
			cardnum,mess[12],idx);
		toneg_start(v->toneg[idx], mess);
		break;

	    case PC_TONEG_RESET:
//			mprintf("[%d/%d] Reseting Tone generation\n",cardnum,ch);
		pThis = (OpenPri *)v->hostdsp;
		toneg = pThis->toneg;
		toneg_reset(v->toneg[idx], mess);
		break;

	    case PC_CODEC_BREAK:
		mprintf("[%d/-]We dont FLASH on PRI\n",cardnum);
		break;

	    case PC_LOOPBACK_ON:
		assert(mess[0] == PC_LLOOPBACK_ON);
		chans[idx].tx_fifo = v->txdf[ch];
		mprintf("[%d/%02d] LoopBack On!\n",cardnum,ch);
		break;

	    case PC_LOOPBACK_OFF:
		assert(mess[0] == PC_LLOOPBACK_OFF);
		chans[idx].tx_fifo = NULL;
		mprintf("[%d/%02d] LoopBack Off!\n",cardnum,ch);
		break;

	    case PC_HOSTECHO_ON:
		assert(mess[0] == PC_LHOSTECHO_ON);
		if (chans[idx].echo_enable == 0){
			chans[idx].echo_enable = 1;
			mprintf("[%d/%02d] Echo cancelor enabled \n",cardnum,ch);
		}
		break;

	    case PC_HOSTECHO_OFF:
		assert(mess[0] == PC_LHOSTECHO_OFF);
		if (chans[idx].echo_enable > 0){
			chans[idx].echo_enable = 0;
			mprintf("[%d/%02d] Echo cancelor disabled \n",cardnum,ch);
		}
		break;

	    case PC_HOSTECHO_OPT:
		assert(mess[0] == PC_LHOSTECHO_OFF);
		/* none at the moment! */
		mprintf("[%d/%02d] Echo cancelor - no optimization at the moment \n",
			cardnum,ch);
		break;

/*
	    case PC_WD_ENABLE:
		assert(mess[0] == PC_LWD_ENABLE);
		if (mess[2]) {
			pci_wd_en_h(pci, i);
		} else {
			pci_wd_en_l(pci, i);
		}
			
		break;

	    case PC_WD_RESET:
		assert(mess[0] == PC_LWD_RESET);
		pci_wd_reset(pci, i);	
		break;

*/
	    case PC_TONED_ST:
		// call progress detector messages 
		toned_add_tone(v->toned[ch], mess);
		break;

	    case PC_TONED_ST_DEL:
		toned_del_tone(v->toned[ch], mess);
		break;

/*
	    case PC_CODEC_GEN:  	
		// general purpose word sent to codec config reg
		id = offset + mess[2];
		//printf("set codec reg ch %d  reg 0x%x val 0x%x\n", id, mess[3], mess[4]);
		hdaports[id].set_codec_reg(mess[3], mess[4]);
		break;
*/

	    case PC_TONED_DEBUG:
		assert(mess[0] == PC_LTONED_DEBUG);
		toned_start_logging(v->toned[ch], mess);
		break;

	    case PC_VOX_UPDATE:
		assert(mess[0] == PC_LVOX_UPDATE);
		if(chans[idx].vox_states != NULL){
			hostvox_update_levels(chans[idx].vox_states, mess);
			mprintf("[%d/-]Updating vox information\n",cardnum);
		}
		else {
			mprintf("[%d/-]No vox_states !\n",cardnum);
		}
		break;

	    case PC_ISDN_CALL:
		mprintf("[%d/%d] Making an ISDN call!\n",
			cardnum,chans[idx].chan_num);
		ret = make_isdn_call(mess,v);
		if(ret){
			rmess[0] = DSP_LISDN_CALL_FAIL;
			rmess[1] = DSP_ISDN_CALL_FAIL;
			rmess[2] = chans[idx].chan_num;
			rmess[3] = ret;
			mprintf("[%d/%d] ISDN CALL Failed (%d)\n",
				cardnum,chans[idx].chan_num,ret);
			v->upmess->Write(rmess, rmess[0]);
		}
		break;

	    case PC_ISDN_GET_CINFO:
		mprintf("[%d/%d]Getting Calling Number\n",
			cardnum,chans[idx].chan_num);
		rmess[0] = DSP_LISDN_GET_CINFO;
		rmess[1] = DSP_ISDN_GET_CINFO;
		rmess[2] = chans[idx].chan_num;
		mprintf("idx(%d) cd(%d) ch(%d) m3(%d)m4(%d)m5(%d)\n",
			idx,cardnum,chans[idx].chan_num,mess[3],mess[4],mess[5]);
		memcpy(&tmpp,&mess[4],mess[3]);
		if ((tmpp != NULL)&&(chans[idx].cid != NULL)){
			memcpy(tmpp,chans[idx].cid,sizeof(isdncid));
		}
//			mprintf("[%d/-]Getting Calling Number: cnum(%d)\n",cardnum,tchp);
//			strcpy(tchp,chans[idx].cid->callingnum);
//			strcpy((char *)mess[3],chans[idx].cid->callingnum);
		v->upmess->Write(rmess, rmess[0]);
		break;

	    default:
		mprintf("[%d/-]Unknown DSP message [%d]\n",cardnum, mess[1]);
		//assert(0);
		break;
	}
} //}}}

static void mprint_pri_message( struct pri* /*pri*/, char *msg )
{
	mprintf("%s\n", msg);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priRunDchannel
	AUTHOR...: Ben Kramer
	DATE.....: 19/11/04
	
	Runs the libpri library interface

\*--------------------------------------------------------------------------*/

void priRunDchannel(void *d)
{ //{{{
	int dfd = 0;
	int swtype = PRI_SWITCH_EUROISDN_E1;
	int node = PRI_CPE;
        //struct wan_sockaddr_ll  sa;
	int d_chan=-1;

	pri_event *e;
	struct timeval tv = {0,0}, *next, ts;
	fd_set rfds, efds;
	int res, tmp;
	int x = 0;
	struct pri *pri;

	word	rmess[COMM_MAX_MESSDSPPC];

	OpenPri* pThis = (OpenPri*)d;
	VPBREG *v      = pThis->vr;
	int cardnum    = v->cardnum;
	int numch      = v->numch;

	OP_CARD *cardinfo = (OP_CARD *)v->cardinfo;

	pri_chan *chans;
	chans = (pri_chan *)v->chans;
	// Initialise channel states
	for(; x < numch; ++x) {
		chans[x].state = CH_IDLE;
		chans[x].run   = 0;
	}

	// Set switch type from config file
	swtype = cardinfo->protocol;
	node = cardinfo->node;
	if (node == ISDN_NODE_LOG) {
		priRunDchannelLog(d);
		return;
	}

	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Get D-channel from config file
	d_chan = cardinfo->d_channel;

	dfd = wanconnect((char*)cardinfo->iface[d_chan].card,
			 (char*)cardinfo->iface[d_chan].name);
	if (dfd < 0){
		mprintf("[%d/-] Failed[%d] to open dchannel '%s'\n",
			cardnum, dfd, cardinfo->iface[d_chan].name);
		goto hell;
	} else {
		mprintf("[%d/-] Opened D-Channel on fd[%d] using '%s'\n",
			cardnum,dfd,cardinfo->iface[d_chan].name);
	}

	// Setup the pri logging to use mprint
	pri_set_message(mprint_pri_message);
	pri_set_error(mprint_pri_message);

	if( !pThis->finish ) {
		pri = pri_new(dfd, node, swtype);
		if (!pri) {
			mprintf("[%d/-] Unable to create PRI\n",cardnum);
			goto hell2;
		}
		pThis->pri = pri;
		pri_set_debug(pri, cardinfo->priverbose);
		mprintf("[%d/-] Created PRI [%p]\n", cardnum, pri);
	}
	while( !pThis->finish ){

		/* Run the D-Channel */
		FD_ZERO(&rfds);
		FD_ZERO(&efds);
		FD_SET(dfd, &rfds);
		FD_SET(dfd, &efds);

		pri = pThis->pri;

		if ((next = pri_schedule_next(pri))) {
			gettimeofday(&tv, NULL);
			tv.tv_sec = next->tv_sec - tv.tv_sec;
			tv.tv_usec = next->tv_usec - tv.tv_usec;
			if (tv.tv_usec < 0) {
				tv.tv_usec += 1000000;
				tv.tv_sec -= 1;
			}
			if (tv.tv_sec < 0) {
				tv.tv_sec = 0;
				tv.tv_usec = 0;
			}
		}
		res = select(dfd + 1, &rfds, NULL, &efds, next ? &tv : NULL);
		e = NULL;

		if (!res) {
			e = pri_schedule_run(pri);
		} else if (res > 0) {
		//	mprintf("[%d/-] PRI check event\n",cardnum );
			e = pri_check_event(pri);
		}

		gettimeofday(&ts, NULL);
		if (e) {
			mprintf("[%d/-] Got an event[%d] on PRI interface at [%ld]...\n",
				cardnum ,e->e, ts.tv_sec);
			handle_pri_event(pri, e, v);
		}
		//else mprintf("[%ld]No events, checking channel statii...\n",ts.tv_sec);

		for(x=0; x < numch; ++x) {
			//mprintf("Checking [%d] state[%d]\n",x,chans[x].state);
			tmp = chans[x].state;
			switch(tmp){
			    case CH_NEEDHANG:
				chans[x].state = CH_IDLE;
				/* Stop Audio Thread */
				chans[x].run=0;
				/* Clear out caller ID information */
				delete chans[x].cid;
				chans[x].cid = NULL;

				res = pri_hangup(pri, chans[x].call,
						 PRI_CAUSE_NORMAL_CLEARING);
				if (res != 0){
					mprintf("[%d/%d] Error hanging up call\n",
						cardnum,chans[x].chan_num);
				}
				// generate event
				if( chans[x].cause != PRI_CAUSE_NORMAL_CLEARING
				 && chans[x].cause != -1 )
				{
					mprintf("[%d/%02d] Sending ISDN busy, cause [%d]..\n",
						cardnum,chans[x].chan_num,chans[x].cause);
					rmess[0] = DSP_LISDN_BUSY;
					rmess[1] = DSP_ISDN_BUSY;
					rmess[2] = chans[x].chan_num;
					rmess[3] = chans[x].cause;
					v->upmess->Write(rmess, rmess[0]);
				}
				mprintf("[%d/%02d] Sending drop event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LDROP;
				rmess[1] = DSP_DROP;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				/* Destroy the call */
				pri_destroycall(pri,chans[x].call);
				chans[x].call = NULL;
				break;

			    case CH_NEEDANS:
				chans[x].run=1;
				//pthread_create(&audio_thread_t[x], NULL,
				//		 audio_thread,(void*)&chans[x]);
				chans[x].state = CH_CONNECTED;
				// generate event
				mprintf("[%d/%02d] Sending isdn answer event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LISDN_ANS;
				rmess[1] = DSP_ISDN_ANS;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				break;

			    case CH_PROCEEDING:
				chans[x].state = CH_SETUP;
				// generate event
				mprintf("[%d/%02d] Sending isdn proceeding event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LISDN_PROCEEDING;
				rmess[1] = DSP_ISDN_PROCEEDING;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				break;

			    case CH_RINGING:
				chans[x].state = CH_SETUP;
				// generate event
				mprintf("[%d/%02d] Sending isdn ringing event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LISDN_RINGING;
				rmess[1] = DSP_ISDN_RINGING;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				break;

			    case CH_NEEDRING:
				mprintf("[%d/%02d] Sending ring event up...(cref:%d)(%s %s %s %s)\n",
					cardnum,
					chans[x].chan_num,
					chans[x].cref,
					chans[x].cid->callingnum,
					chans[x].cid->callingname,
					chans[x].cid->callednum,
					chans[x].cid->calledsubaddr);
				rmess[0] = DSP_LISDN_RING;
				rmess[1] = DSP_ISDN_RING;
				rmess[2] = chans[x].chan_num;
				chans[x].state=CH_RING;
				v->upmess->Write(rmess, rmess[0]);
				break;
			}
		}
	}

    hell2:
	mprintf("[%d/-] Closing socket to D-channel\n",cardnum);
	close(dfd);
	dfd = -1;

    hell:
	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	pthread_detach(pthread_self());
	mprintf("[%d/-] Finishing priRunDchannel\n",cardnum);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: priRunDchannelLog
	AUTHOR...: Ben Kramer
	DATE.....: 19/11/04
	
	Runs the libpri library on 2 interfaces for logging and combines the
	results.

\*--------------------------------------------------------------------------*/

void priRunDchannelLog(void *d)
{
	int swtype = PRI_SWITCH_EUROISDN_E1;
	int node = PRI_CPE;
        //struct wan_sockaddr_ll  sa;
	int d_chan=-1;

	pri_event *e;
	struct timeval ts;
	#ifndef CASLOG
	int dfd;
	int dfd2;
	#endif
	int seldfd;
	fd_set rfds, efds;
	int res, i, x = 0;
	struct pri *pri;
	struct pri *pri2;
	struct pri *pritmp;
	int islog=0;

	char *tmpstr;

	struct cref_info *cref_info;
	struct cref_info *cref_info_last;

	word	rmess[COMM_MAX_MESSDSPPC];

	OpenPri* pThis = (OpenPri*)d;
	VPBREG *v      = pThis->vr;
	int cardnum    = v->cardnum;
	int numch      = v->numch;

	OP_CARD *cardinfo = (OP_CARD *)v->cardinfo;

	pri_chan *chans;
	chans = (pri_chan *)v->chans;

	// Initialise channel states
	for(; x < numch; ++x) {
		chans[x].state = CH_IDLE;
	    #ifdef CASLOG
		chans[x].run = 1;
	    #else
		chans[x].run = 0;
	    #endif
	}

	// Set switch type from config file
	swtype = cardinfo->protocol;
	node = cardinfo->node;
	if (node == ISDN_NODE_LOG) {
		node = PRI_CPE;
		islog=1;
	} else {
		priRunDchannel(d);
		return;
	}

	GenericEnterCriticalSection(&pThis->mutex);
	++pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	// Split up channel and card names
	// iface is initialised from config file which uses channel1 - channel31
	// when numch = 30.
	// NOTE: bad things will happen here if any (active) channel is not listed
	//       in the config file.  FIXME
	for(i=1; i < numch + 2; ++i){
		tmpstr = strchr((char *)cardinfo->iface[i].name,',');
		if (tmpstr != NULL){
			*tmpstr = '\0';
			tmpstr++;
			strcpy((char *)cardinfo->iface[i].name2,tmpstr);
		}
		tmpstr = strchr((char *)cardinfo->iface[i].card,',');
		if (tmpstr != NULL){
			*tmpstr = '\0';
			tmpstr++;
			strcpy((char *)cardinfo->iface[i].card2,tmpstr);
		}
	}
	// Get D-channel from config file
	d_chan = cardinfo->d_channel;

	#ifndef CASLOG
	dfd = wanconnect((char*)cardinfo->iface[d_chan].card,
			 (char*)cardinfo->iface[d_chan].name);
	dfd2 = wanconnect((char*)cardinfo->iface[d_chan].card2,
			  (char*)cardinfo->iface[d_chan].name2);
	if( dfd < 0 || dfd2 < 0 ){
		mprintf("[%d/-] Failed[%d] to open dchannel '%s'\n",
			cardnum, dfd, cardinfo->iface[d_chan].name);
		goto hell;
	} else {
		mprintf("[%d/-] Opened 1/2 D-Channel on fd[%d] using '%s'\n",
			cardnum,dfd,cardinfo->iface[d_chan].name);
		mprintf("[%d/-] Opened 1/2 D-Channel on fd[%d] using '%s'\n",
			cardnum,dfd2,cardinfo->iface[d_chan].name2);
	}
	mprintf("[%d/-] D-Channel is in logging mode!\n",cardnum);
	seldfd = dfd2 > dfd ? dfd2 : dfd;

	// Setup the pri logging to use mprint
	pri_set_message(mprint_pri_message);
	pri_set_error(mprint_pri_message);

	if( !pThis->finish ) {
		pri = pri_new(dfd, node, swtype);
		if( !pri ) {
			mprintf("[%d/-] Unable to create PRI (1)\n",cardnum);
			goto hell2;
		}
		pri2 = pri_new(dfd2, node, swtype);
		if( !pri2 ) {
			mprintf("[%d/-] Unable to create PRI (2)\n",cardnum);
			free( pri );
			goto hell2;
		}
		pThis->pri = pri;
		pThis->pri2 = pri2;
		pri_set_debug(pri, cardinfo->priverbose);
		pri_set_debug(pri2, cardinfo->priverbose);
		mprintf("[%d/-] Created PRI [%p]\n",cardnum,pri);
	}
	while (!pThis->finish){

		struct timeval tv = {1,0};

		/* Run the D-Channel */
		FD_ZERO(&rfds);
		FD_ZERO(&efds);
		FD_SET(dfd, &rfds);
		FD_SET(dfd, &efds);
		FD_SET(dfd2, &rfds);
		FD_SET(dfd2, &efds);

		pri = pThis->pri;
		pri2 = pThis->pri2;
		res = select(seldfd + 1, &rfds, NULL, &efds, &tv);
		if( res <= 0 ) continue;

		//mprintf("[%d/-] PRI check event\n",cardnum );
		e = NULL;
		if (FD_ISSET(dfd,&rfds)){
			e = pri_check_event_log(pri);
			pritmp = pri;
		} else if (FD_ISSET(dfd2,&rfds)){
			e = pri_check_event_log(pri2);
			pritmp = pri2;
		}

		gettimeofday(&ts, NULL);
		if (e) {
			handle_pri_event_log(pritmp, e, d);
		}
		// else mprintf("[%ld]No events, checking channel statii...\n",ts.tv_sec);

		for (x=0; x < v->numch; ++x) {
			//mprintf("Checking [%d] state[%d]\n",x,chans[x].state);
			switch(chans[x].state){
			case CH_NEEDHANG:
				chans[x].state = CH_IDLE;
				/* Stop Audio Thread */
				chans[x].run=0;

				/* Clear out caller ID information */
				delete chans[x].cid;
				chans[x].cid = NULL;

				// generate event
				if (chans[x].cause != PRI_CAUSE_NORMAL_CLEARING){
					mprintf("[%d/%02d] Sending ISDN busy, cause [%d]..\n",
						cardnum,chans[x].chan_num,chans[x].cause);
					rmess[0] = DSP_LISDN_BUSY;
					rmess[1] = DSP_ISDN_BUSY;
					rmess[2] = chans[x].chan_num;
					rmess[3] = chans[x].cause;
					v->upmess->Write(rmess, rmess[0]);
				}
				mprintf("[%d/%02d] Sending drop event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LDROP;
				rmess[1] = DSP_DROP;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				//pri_destroycall(pri,chans[x].call);
				pri_destroycall(pritmp, e->hangup.call);
				chans[x].call = NULL;
				chans[x].cref = -1;
				chans[x].cref_check = 0;
				break;

			case CH_NEEDANS:
				chans[x].run=1;
				//pthread_create(&audio_thread_t[x], NULL,
				//		 audio_thread_log,(void*)&chans[x]);
				chans[x].state = CH_CONNECTED;
				// generate event
				if (chans[x].cref_check == 0){
					mprintf("[%d/%02d] checking cref_info linked list for 0x%X\n",
						cardnum,chans[x].chan_num, chans[x].cref);
					if (chans[x].cref != -1){
						cref_info = (struct cref_info*)pThis->cref_info;
						while(cref_info != NULL){
							mprintf("[%d/%02d] comparing cref_info  0x%X to 0x%X (0x%X)\n",
								cardnum,chans[x].chan_num,
								cref_info->cref, chans[x].cref,
								(chans[x].cref&0x7fff));

							if(cref_info->cref == (chans[x].cref&0x7fff)){
								mprintf("[%d/%02d] Found call information from cref_info linked list\n",
									cardnum,chans[x].chan_num);
								if (chans[x].cid == NULL){
									chans[x].cid = new isdncid;
								}
								chans[x].cid->callingplan = cref_info->cid.callingplan;
								chans[x].cid->callingpres = cref_info->cid.callingpres;
								chans[x].cid->calledplan = cref_info->cid.calledplan;
								strcpy(chans[x].cid->callingnum,cref_info->cid.callingnum);
								strcpy(chans[x].cid->callingname,cref_info->cid.callingname);
								strcpy(chans[x].cid->callednum,cref_info->cid.callednum);
								chans[x].cref_check=1;
								chans[x].layer1=cref_info->layer1;
								chans[x].transcap=cref_info->transcap;
								if(cref_info_last != NULL){
									cref_info_last->next = cref_info->next;
								}
								else {
									pThis->cref_info = NULL;
								}
								delete cref_info;
								cref_info = NULL;
							}
							else {
								cref_info_last = cref_info;
								cref_info = (struct cref_info*)cref_info->next;
							}
						}
					}
					else {
						mprintf("[%d/%02d] No caller reference number!\n",
							cardnum,chans[x].chan_num);
					}
				}

				mprintf("[%d/%02d] Sending isdn answer event up...\n",
					cardnum,chans[x].chan_num);
				rmess[0] = DSP_LISDN_ANS;
				rmess[1] = DSP_ISDN_ANS;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				break;

			case CH_NEEDRING:
				mprintf("[%d/%02d] Sending ring event up...(cref:%d)(%s %s %s)\n",
					cardnum,chans[x].chan_num,
					chans[x].cref,chans[x].cid->callingnum,
					chans[x].cid->callingname,chans[x].cid->callednum);
				rmess[0] = DSP_LISDN_RING;
				rmess[1] = DSP_ISDN_RING;
				rmess[2] = chans[x].chan_num;
				v->upmess->Write(rmess, rmess[0]);
				chans[x].state=CH_RING;
				break;

			case CH_NEEDTERMINATING:
				/* Stop Audio Thread */
				chans[x].run=0;
				chans[x].state=CH_TERMINATING;
				break;
			}
		}
	}

    hell2:
	mprintf("[%d/-] Closing socket to D-channel\n",cardnum);
	close(dfd);
	close(dfd2);
	dfd = -1;
	#else
	mprintf("[%d/-] D-Channel is disabled in cas logging mode!\n",cardnum);
	while(!pThis->finish) {
		sleep(1);
	}
	#endif

    hell:
	GenericEnterCriticalSection(&pThis->mutex);
	--pThis->active_threads;
	GenericLeaveCriticalSection(&pThis->mutex);

	pthread_detach(pthread_self());
	mprintf("[%d/-] Finishing priRunDchannel\n",cardnum);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: handle_pri_event
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04
	
	Decides what to do with a PRI event

\*--------------------------------------------------------------------------*/
static void handle_pri_event(struct pri *pri, pri_event *e, VPBREG *v)
{
	pri_chan *chans;
	chans = (pri_chan *)v->chans;
	int idx,i;

	switch(e->e) {
	    case PRI_EVENT_DCHAN_UP:
		mprintf("[%d/-] PRI D-Channel is now up!  :-)\n",v->cardnum);
		break;

	    case PRI_EVENT_DCHAN_DOWN:
		mprintf("[%d/-] PRI D-Channel is now down! :-(\n",v->cardnum);
		break;

	    case PRI_EVENT_RESTART:
		mprintf("[%d/-] PRI Restarting channel %d\n",v->cardnum, e->restart.channel);
//		hangup_channel(e->restart.channel);
		break;

	    case PRI_EVENT_CONFIG_ERR:
		mprintf("[%d/-] PRI Configuration error detected: %s\n",v->cardnum, e->err.err);
		break;

	    case PRI_EVENT_RING:
		mprintf("[%d/%d] PRI Ring on channel (from [%s] to [%s])(cref:%d)\n",
			v->cardnum, e->ring.channel, e->ring.callingnum,
			e->ring.callednum,e->ring.cref);
		if (e->ring.channel != -1){
			try {
				idx = get_idx_from_ich(v,e->ring.channel);
			} catch ( const Wobbly &w ) {
				mprintf("handle_pri_event: EVENT_RING Failed to get channel from %d\n",
					e->ring.channel);
				return;
			}
		}
		else {
			//i = slot2chan(e->ring.slotmap,v->numch);
			//mprintf("[%d/%d]       Slotmap'ed to channel %d\n",v->cardnum, i,i);
			//idx = get_idx_from_ich(v,i);
			idx = findfreechan(v);
			mprintf("[%d/%d]       Got Slotmap, found free channel %d\n",
				v->cardnum,idx,idx);
		}
		chans[idx].state=CH_NEEDRING;
		/* record call informaton */
		chans[idx].call = e->ring.call;
		/* Save caller ID info */
		if (chans[idx].cid == NULL){
			chans[idx].cid = new isdncid;
		}
		chans[idx].cid->callingplan = e->ring.callingplan;
		chans[idx].cid->callingpres = e->ring.callingpres;
		chans[idx].cid->calledplan = e->ring.calledplan;
		strcpy(chans[idx].cid->callingnum,e->ring.callingnum);
		strcpy(chans[idx].cid->callingname,e->ring.callingname);
		strcpy(chans[idx].cid->calledsubaddr,e->ring.calledsubaddr);
		strcpy(chans[idx].cid->callednum,e->ring.callednum);
		chans[idx].cref=e->ring.cref;
		chans[idx].layer1=e->ring.layer1;
		chans[idx].transcap=e->ring.ctype;
		for(i=0;i<e->ring.lowlayercompatibility[0]+1;i++){
			chans[idx].lowlayercompatibility[i]=e->ring.lowlayercompatibility[i];
		}
		break;

	    case PRI_EVENT_HANGUP:
		try {
			idx = get_idx_from_ich(v,e->hangup.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event: EVENT_HANGUP Failed to get channel from %d\n",
				e->hangup.channel);
			return;
		}
		mprintf("[%d/%d] PRI Hangup on channel (cref:%d)\n",
			v->cardnum , idx, e->hangup.cref);
		chans[idx].cause=e->hangup.cause;
		chans[idx].state=CH_NEEDHANG;
		break;

	    case PRI_EVENT_HANGUP_REQ:
		try {
			idx = get_idx_from_ich(v,e->hangup.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event: EVENT_HANGUP_REQ Failed to get channel from %d\n",
				e->hangup.channel);
			return;
		}
		mprintf("[%d/%d] PRI Hangup requested on channel (cref:%d)\n",
			v->cardnum , idx, e->hangup.cref);
		chans[idx].cause=e->hangup.cause;
		chans[idx].state=CH_NEEDHANG;
		break;

	    case PRI_EVENT_RINGING:
		try {
			idx = get_idx_from_ich(v,e->ringing.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event: EVENT_RINGING Failed to get channel from %d\n",
				e->ringing.channel);
			return;
		}
		mprintf("[%d/%d] PRI Got a Ringing on channel (cref:%d)\n",
			v->cardnum, e->ringing.channel, e->ringing.cref);
		chans[idx].state=CH_RINGING;
		break;

	    case PRI_EVENT_PROCEEDING:
		try {
			idx = get_idx_from_ich(v,e->proceeding.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event: EVENT_PROCEEDING Failed to get channel from %d\n",
				e->proceeding.channel);
			return;
		}
		mprintf("[%d/%d] PRI Got a Call Proceeding on channel \n",
			v->cardnum , e->proceeding.channel);
		chans[idx].state=CH_PROCEEDING;
		break;

	    case PRI_EVENT_ANSWER:
		mprintf("[%d/%d] PRI Got a Answer on channel (cref:%d)\n",
			v->cardnum , e->answer.channel,e->answer.cref);
		try {
			idx = get_idx_from_ich(v,e->answer.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event: EVENT_ANSWER Failed to get channel from %d\n",
				e->answer.channel);
			return;
		}
		if (chans[idx].state == CH_SETUP){
			chans[idx].state=CH_NEEDANS;
		}
		else {
			mprintf("[%d/%d] PRI Channel not in setup state, in %d, hanging up\n",
				v->cardnum , e->answer.channel,chans[e->ring.channel].state);
			chans[idx].state=CH_NEEDHANG;
		}
		break;

	    case PRI_EVENT_HANGUP_ACK:
		mprintf("[%d/%d] PRI Hangup ACK on channel (cref:%d)\n",
			v->cardnum , e->hangup.channel,e->hangup.cref);
		/* Ignore */
		break;

	    case PRI_EVENT_INFO_RECEIVED:
		mprintf("[%d/%d] PRI number is: %s (cref:%d)\n",
			v->cardnum, e->ring.channel , e->ring.callednum,e->ring.cref);
		if(!(strlen(e->ring.callednum) < 3)){
			mprintf("[%d/%d] PRI Final number is: %s\n",
				v->cardnum, e->ring.channel, e->ring.callednum);
			pri_answer(pri, e->ring.call, e->ring.channel, 1);
		}
		break;

	    default:
		mprintf("[%d/-] PRI Unknown PRI event [%d] !!\n",v->cardnum , e->e);
	}
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: handle_pri_event_log
	AUTHOR...: Ben Kramer
	DATE.....: 14/07/2005
	
	Decides what to do with a PRI event

\*--------------------------------------------------------------------------*/
static void handle_pri_event_log(struct pri *pri, pri_event *e, void *d)
{
	OpenPri* pThis = (OpenPri *)d;
	VPBREG *v = pThis->vr;
	struct cref_info *cref_info;

	pri_chan *chans;
	chans = (pri_chan *)v->chans;
	int idx;

	switch(e->e) {
	    case PRI_EVENT_DCHAN_UP:
//		mprintf("[%d/-] PRI D-Channel is now up!  :-)\n",v->cardnum);
		break;

	    case PRI_EVENT_DCHAN_DOWN:
//		mprintf("[%d/-] PRI D-Channel is now down! :-(\n",v->cardnum);
		break;

	    case PRI_EVENT_RESTART:
//		mprintf("[%d/-] PRI Restarting channel %d\n",v->cardnum, e->restart.channel);
//		hangup_channel(e->restart.channel);
		break;

	    case PRI_EVENT_CONFIG_ERR:
		mprintf("[%d/-] PRI Configuration error detected: %s\n",
			v->cardnum, e->err.err);
		break;

	    case PRI_EVENT_RING:
		mprintf("[%d/%d] PRI Ring on channel (from [%s] to [%s])(cref:%d)\n",
			v->cardnum, e->ring.channel, e->ring.callingnum,
			e->ring.callednum,e->ring.cref);
		if(e->ring.channel != -1){
			try {
				idx = get_idx_from_ich(v,e->ring.channel);
			} catch ( const Wobbly &w ) {
				mprintf("handle_pri_event_log: EVENT_RING Failed to get channel from %d\n",
					e->ring.channel);
				return;
			}
			chans[idx].state=CH_NEEDRING;
			/* record call informaton */
			chans[idx].call = e->ring.call;
			/* Save caller ID info */
			if (chans[idx].cid == NULL){
				chans[idx].cid = new isdncid;
			}
			chans[idx].cid->callingplan = e->ring.callingplan;
			chans[idx].cid->callingpres = e->ring.callingpres;
			chans[idx].cid->calledplan = e->ring.calledplan;
			strcpy(chans[idx].cid->callingnum,e->ring.callingnum);
			strcpy(chans[idx].cid->callingname,e->ring.callingname);
			strcpy(chans[idx].cid->callednum,e->ring.callednum);
			chans[idx].cref=e->ring.cref;
			chans[idx].cref_check=1;
			chans[idx].layer1=e->ring.layer1;
			chans[idx].transcap=e->ring.ctype;
		}
		else {
			cref_info = (struct cref_info*)pThis->cref_info;
			if(cref_info != NULL){
				while(cref_info->next != NULL)
					cref_info = (struct cref_info*)cref_info->next;
				cref_info->next = (void *)new struct cref_info;
				cref_info = (struct cref_info*)cref_info->next;
			}
			else {
				cref_info = new struct cref_info;
				pThis->cref_info = (void *)cref_info;
			}
			cref_info->next = NULL;
			cref_info->cid.callingplan = e->ring.callingplan;
			cref_info->cid.callingpres = e->ring.callingpres;
			cref_info->cid.calledplan = e->ring.calledplan;
			strcpy(cref_info->cid.callingnum,e->ring.callingnum);
			strcpy(cref_info->cid.callingname,e->ring.callingname);
			strcpy(cref_info->cid.callednum,e->ring.callednum);
			cref_info->cref=e->ring.cref;
			cref_info->layer1=e->ring.layer1;
			cref_info->transcap=e->ring.ctype;
		}
		break;

	    case PRI_EVENT_HANGUP:
		mprintf("[%d/%d] PRI Hangup on channel (cref:%d)\n",
			v->cardnum , e->hangup.channel, e->hangup.cref);
		try {
			idx = get_idx_from_ich(v,e->hangup.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event_log: EVENT_HANGUP Failed to get channel from %d\n",
				e->hangup.channel);
			q931_destroycall(pri, e->hangup.cref);
			return;
		}
		chans[idx].run=0;
		chans[idx].cause=e->hangup.cause;
		chans[idx].state=CH_NEEDHANG;
		break;

	    case PRI_EVENT_HANGUP_REQ:
		mprintf("[%d/%d] PRI Hangup requested on channel (cref:%d)\n",
			v->cardnum , e->hangup.channel, e->hangup.cref);
		try {
			idx = get_idx_from_ich(v,e->hangup.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event_log: EVENT_HANGUP_REQ Failed to get channel from %d\n",
				e->hangup.channel);
			q931_destroycall(pri, e->hangup.cref);
			return;
		}
		chans[idx].run=0;
		chans[idx].cause=e->hangup.cause;
		//chans[idx].state=CH_NEEDHANG;
		chans[idx].state=CH_NEEDTERMINATING;
		break;

	    case PRI_EVENT_RINGING:
		mprintf("[%d/%d] PRI Got a Ringing on channel (cref:%d)\n",
			v->cardnum , e->ringing.channel,e->ringing.cref);
		// For UK logger outgoing calls
		if(e->ringing.channel != -1){
			try {
				idx = get_idx_from_ich(v,e->ringing.channel);
			} catch ( const Wobbly &w ) {
				mprintf("handle_pri_event_log: EVENT_RING Failed to get channel from %d\n",
					e->ringing.channel);
				return;
			}
			if(chans[idx].state==CH_RING) {
				  // ok... already in CH_RING
				mprintf("[%d/%d] PRI Ringing : already in CH_RING \n",
					v->cardnum , e->ringing.channel);

			}else{

					chans[idx].state=CH_NEEDRING;
   					mprintf("[%d/%d] PRI Ringing : added a CH_NEEDRING \n",
						v->cardnum , e->ringing.channel);

					/* record call informaton */
					chans[idx].call = e->ringing.call;
					/* Save caller ID info */
					if (chans[idx].cid == NULL){
						chans[idx].cid = new isdncid;
					}
					chans[idx].cid->callingplan = e->ring.callingplan;
					chans[idx].cid->callingpres = e->ring.callingpres;
					chans[idx].cid->calledplan = e->ring.calledplan;
					strcpy(chans[idx].cid->callingnum,e->ring.callingnum);
					strcpy(chans[idx].cid->callingname,e->ring.callingname);

					chans[idx].cref=e->ringing.cref;
					chans[idx].cref_check=1;
					chans[idx].layer1=e->ring.layer1;
					chans[idx].transcap=e->ring.ctype;
				}

		}

		break;

	    case PRI_EVENT_PROCEEDING:
		mprintf("[%d/%d] PRI Got a Call Proceeding on channel \n",
			v->cardnum , e->proceeding.channel);

		if(e->proceeding.channel != -1){
			try {
				idx = get_idx_from_ich(v,e->proceeding.channel);
			} catch ( const Wobbly &w ) {
				mprintf("handle_pri_event_log: EVENT_RING Failed to get channel from %d\n",
					e->proceeding.channel);
				return;
			}
			if(chans[idx].state==CH_RING) {
				  // ok... already in CH_RING
  				mprintf("[%d/%d] PRI Call Proceeding : already in CH_RING \n",
					v->cardnum , e->proceeding.channel);

			}else{

					chans[idx].state=CH_NEEDRING;
	   				mprintf("[%d/%d] PRI Call Proceeding : added a CH_NEEDRING \n",
						v->cardnum , e->proceeding.channel);

					/* record call informaton */
					chans[idx].call = e->proceeding.call;
					/* Save caller ID info */
					if (chans[idx].cid == NULL){
						chans[idx].cid = new isdncid;
					}
					chans[idx].cid->callingplan = e->ring.callingplan;
					chans[idx].cid->callingpres = e->ring.callingpres;
					chans[idx].cid->calledplan = e->ring.calledplan;
					strcpy(chans[idx].cid->callingnum,e->ring.callingnum);
					strcpy(chans[idx].cid->callingname,e->ring.callingname);

					chans[idx].cref=e->proceeding.cref;
					chans[idx].cref_check=1;
					chans[idx].layer1=e->ring.layer1;
					chans[idx].transcap=e->ring.ctype;
			}
			
		}



		break;

	    case PRI_EVENT_ANSWER:
		mprintf("[%d/%d] PRI Got a Answer on channel (ans.cref:%d)\n",
			v->cardnum , e->answer.channel,e->answer.cref);
		try {
			idx = get_idx_from_ich(v,e->answer.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event_log: EVENT_ANSWER Failed to get channel from %d\n",
				e->answer.channel);
			return;
		}
		chans[idx].state=CH_NEEDANS;
		if (chans[idx].cref != e->answer.cref){
			chans[idx].cref = e->answer.cref;
			mprintf("[%d/%d] Updated cref %d\n",
				v->cardnum , chans[idx].chan_num, chans[idx].cref);
		}

		break;

	    case PRI_EVENT_SETUP_ACK:
		mprintf("[%d/%d] PRI Setup ACK on channel \n",v->cardnum,e->setup_ack.channel);
		break;

	    case PRI_EVENT_HANGUP_ACK:
		mprintf("[%d/%d] PRI Hangup ACK on channel (cref:%d)\n",
			v->cardnum , e->hangup.channel,e->hangup.cref);
		/* Ignore */
		try {
			idx = get_idx_from_ich(v,e->hangup.channel);
		} catch ( const Wobbly &w ) {
			mprintf("handle_pri_event_log: EVENT_HANGUP_ACK Failed to get channel from %d\n",
				e->hangup.channel);
			q931_destroycall(pri, e->hangup.cref);
			return;
		}
		chans[idx].run=0;
		chans[idx].cause=e->hangup.cause;
		chans[idx].state=CH_NEEDHANG;
		break;

	    case PRI_EVENT_INFO_RECEIVED:
		mprintf("[%d/%d] PRI number is: %s (cref:%d)\n",v->cardnum, e->ring.channel , e->ring.callednum,e->ring.cref);
		break;

	    default:
		mprintf("[%d/-] PRI Unknown PRI event [%d] !!\n",v->cardnum , e->e);
	}
}

#if 0
/*--------------------------------------------------------------------------*\

	FUNCTION.: audio_thread
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04
	
	Connects to the correct wanpipe interface and gets/puts the audio onto
	the correct driver FIFO's

\*--------------------------------------------------------------------------*/

static void *audio_thread(void *pv)
{ //{{{
	unsigned char bit_inv[256];
	pri_chan * c;
	c = (pri_chan *)pv;
	//struct wan_sockaddr_ll  sa;
	int c_fd,i;
	int ret,nbytes,err;
	unsigned char buf[16+NBUF];
//	char chint[16];
	int connected=0;
	int encoding;

//	api_rx_element_t* api_rx_el;

	// Select stuff
	fd_set port_fds,write_fds,oob_fds;
	struct timeval tv;

	pthread_detach(pthread_self());
	for(i=0;i<256;i++){
		bit_inv[i] = (i&1?0x80:0);
		bit_inv[i] |= (i&2?0x40:0);
		bit_inv[i] |= (i&4?0x20:0);
		bit_inv[i] |= (i&8?0x10:0);
		bit_inv[i] |= (i&0x10?0x08:0);
		bit_inv[i] |= (i&0x20?0x04:0);
		bit_inv[i] |= (i&0x40?0x02:0);
		bit_inv[i] |= (i&0x80?0x01:0);
	}

	// check encoding
	if (((c->layer1 == PRI_LAYER_1_ULAW)&&(c->encoding != VPB_MULAW)) // Compare if mulaw
	   || ((c->layer1 == PRI_LAYER_1_ALAW)&&(c->encoding != VPB_ALAW)) // Compare if alaw
	   ){
		mprintf("Fixing Layer1(0x%02X) and encoding(%d)!\n",c->layer1,c->encoding);
		// Encoding doesnt match layer1 from call setup, fixit!
		if (c->layer1 == PRI_LAYER_1_ULAW){
			encoding = VPB_MULAW;
		} else if (c->layer1 == PRI_LAYER_1_ALAW){
			encoding = VPB_ALAW;
		} else {
			encoding = VPB_RAW;
		}
	} else {
		encoding = c->encoding;
	}
	if (c->transcap == PRI_TRANS_CAP_DIGITAL){
		encoding = VPB_RAW;
	}
	c->run = 1;
	mprintf("Starting audio on channel [%d] interface [%s] encoding[%s]"
		" layer1(0x%02X) transcap(0x%02X)\n",
		c->chan_num,c->iface->name,
		(VPB_ALAW==encoding ? "alaw"
				    : (VPB_MULAW==encoding ? "mulaw" : "raw")),
		c->layer1, c->transcap);
	c_fd = wanconnect((char *)c->iface->card,(char *)c->iface->name);
        if (c_fd < 0) {
                mprintf("Failed to connect to channel interface on card '%s' iface '%s': %d\n",
			c->iface->card,c->iface->name, c_fd);
		pthread_exit((void*)0);
        }
        else {
                mprintf("Connected to channel[%d] interface[%s] socket on fd[%d]\n",
			c->chan_num,c->iface->name,c_fd);
		connected=1;
        }
	c->v->rxdf[c->chan_num]->SetState(1);
	c->v->txdf[c->chan_num]->SetState(1);
	c->v->rxdf[c->chan_num]->Flush();
	c->v->txdf[c->chan_num]->Flush();
	while ((c->run ==1)&&(connected)){
		/* Set up file descriptor set for select()ing on sockets */
		FD_ZERO(&port_fds);
		FD_ZERO(&write_fds);
		FD_ZERO(&oob_fds);
		FD_SET(c_fd,&port_fds);
		FD_SET(c_fd,&write_fds);
		FD_SET(c_fd,&oob_fds);
		/* wait for a socket to be ready */
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		ret = select(c_fd+1,&port_fds,NULL,&oob_fds,&tv);
		if (ret > 0){
			if (FD_ISSET(c_fd,&oob_fds)){
				nbytes = recv(c_fd, buf, sizeof(buf), MSG_OOB);
				if(nbytes < 0 ) {
					mprintf("audio_thread: Failed to receive OOB \n");
					err = ioctl(c_fd,SIOC_WANPIPE_SOCK_STATE,0);
					mprintf("audio_thread: Sock state is %s\n",
						(err == 0) ? "CONNECTED" :
						(err == 1) ? "DISCONNECTED" :
						"CONNECTING");
					if (err == 1){
						mprintf("audio_thread: Terminating due to disconnection\n");
						connected =0;
					}
				}
			}
			if (FD_ISSET(c_fd,&port_fds)){
				/* Get data from socket - record */
				if ((nbytes = recv(c_fd, buf, sizeof(buf), 0)) <= 0) {
					// got error or connection closed by client
					if (nbytes == 0) {
						// connection closed
						mprintf("[%s][%d] Socket hangup ...\n",
							c->iface->card, c->iface->name);
					} else {
						mprintf("[%s][%d] Socket error[%d] ...\n",
							c->iface->card, c->iface->name,errno);
					}
				} else {
					for(i=0;i<(nbytes-16);i++){
						buf[16+i] = bit_inv[buf[16+i]];
					}
					/* Send data up FIFO - record  */
					prepare_rec_buf(&buf[16], c->v, c->chan_num,encoding);
					/* Get data from FIFO - play */
					prepare_play_buf(&buf[16], c->v, c->chan_num,encoding);
					for(i=0;i<(nbytes-16);i++){
						buf[16+i] = bit_inv[buf[16+i]];
					}
					/* Send FIFO up socket  - play */
					err = send(c_fd,buf,sizeof(buf),0);
				}
			}
			nbytes=0;
		}
	}
	c->v->rxdf[c->chan_num]->SetState(0);
	c->v->txdf[c->chan_num]->SetState(0);
	close(c_fd);
	mprintf("Finished audio on channel [%d] interface [%s] encoding[%s]\n",
		c->chan_num,c->iface->name,
		(VPB_ALAW==encoding ? "alaw"
				    : (VPB_MULAW==encoding ? "mulaw" : "raw")));
	pthread_exit((void*)0);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: audio_thread_log
	AUTHOR...: Ben Kramer
	DATE.....: 19/07/05
	
	Connects to the correct wanpipe interfaces and gets/puts the audio onto
	the correct driver FIFO's after combing them from both interfaces!

\*--------------------------------------------------------------------------*/

static void *audio_thread_log(void *pv)
{ //{{{
	unsigned char bit_inv[256];
	pri_chan * c;
	c = (pri_chan *)pv;
	//struct wan_sockaddr_ll  sa;
	int i;
	int c_fd,c_fd2,selfd;
	int ret,nbytes,err;
	unsigned char buf[16+NBUF];
	unsigned char buf2[16+NBUF];
//	char chint[16];
	int connected=0;
	int encoding;

//	api_rx_element_t* api_rx_el;

	// Select stuff
	fd_set port_fds,write_fds,oob_fds;
	struct timeval tv;

	struct timeval logtime;

	pthread_detach(pthread_self());
	for(i=0;i<256;i++){
		bit_inv[i] = (i&1?0x80:0);
		bit_inv[i] |= (i&2?0x40:0);
		bit_inv[i] |= (i&4?0x20:0);
		bit_inv[i] |= (i&8?0x10:0);
		bit_inv[i] |= (i&0x10?0x08:0);
		bit_inv[i] |= (i&0x20?0x04:0);
		bit_inv[i] |= (i&0x40?0x02:0);
		bit_inv[i] |= (i&0x80?0x01:0);
	}

	// check encoding
	if (((c->layer1 == PRI_LAYER_1_ULAW)&&(c->encoding != VPB_MULAW)) // Compare if mulaw
	   || ((c->layer1 == PRI_LAYER_1_ALAW)&&(c->encoding != VPB_ALAW)) // Compare if alaw
	   ){
		// Encoding doesnt match layer1 from call setup, fixit!
		if (c->layer1 == PRI_LAYER_1_ULAW){
			encoding = VPB_MULAW;
		} else if (c->layer1 == PRI_LAYER_1_ALAW){
			encoding = VPB_ALAW;
		} else {
			encoding = VPB_RAW;
		}
	} else {
		encoding = c->encoding;
	}
	c->run = 1;
	gettimeofday(&logtime,NULL);
	mprintf("[%ld:%ld]Starting LOGGING audio on channel [%d] interface [%s][%s] encoding[%s]\n",
		logtime.tv_sec,logtime.tv_usec,c->chan_num,c->iface->name, c->iface->name2,
		(VPB_ALAW==encoding ? "alaw" : (VPB_MULAW==encoding ? "mulaw" : "raw")));
	c_fd = wanconnect((char *)c->iface->card,(char *)c->iface->name);
	c_fd2 = wanconnect((char *)c->iface->card2,(char *)c->iface->name2);
        if (c_fd < 0) {
                mprintf("Failed to connect to channel interface on card '%s' iface '%s': %d\n",
			c->iface->card,c->iface->name, c_fd);
		pthread_exit((void*)0);
        } else if (c_fd2 < 0) {
                mprintf("Failed to connect to channel interface on card '%s' iface '%s': %d\n",
			c->iface->card2,c->iface->name2, c_fd2);
		pthread_exit((void*)0);
        }
        else {
                mprintf("Connected to channel[%d] interface[%s] socket on fd[%d]\n",
			c->chan_num,c->iface->name,c_fd);
                mprintf("Connected to channel[%d] interface[%s] socket on fd[%d]\n",
			c->chan_num,c->iface->name2,c_fd2);
		connected=1;
        }
	if (c_fd2>c_fd){
		selfd=c_fd2;
	} else {
		selfd=c_fd;
	}
	c->v->rxdf[c->chan_num]->SetState(1);
	c->v->txdf[c->chan_num]->SetState(1);
	while ((c->run ==1)&&(connected)){
		/* Set up file descriptor set for select()ing on sockets */
		FD_ZERO(&port_fds);
		FD_ZERO(&write_fds);
		FD_ZERO(&oob_fds);
		FD_SET(c_fd,&port_fds);
		FD_SET(c_fd,&write_fds);
		FD_SET(c_fd,&oob_fds);
//		FD_SET(c_fd2,&port_fds);
//		FD_SET(c_fd2,&write_fds);
//		FD_SET(c_fd2,&oob_fds);
		/* wait for a socket to be ready */
		tv.tv_sec = 1;
		tv.tv_usec = 0;
//		ret = select(selfd+1,&port_fds,NULL,&oob_fds,&tv);
		ret = select(c_fd+1,&port_fds,NULL,&oob_fds,&tv);
		if (ret > 0){
			if (FD_ISSET(c_fd,&oob_fds)){
				nbytes = recv(c_fd, buf, sizeof(buf), MSG_OOB);
				if(nbytes < 0 ) {
					mprintf("audio_thread: Failed to receive OOB \n");
					err = ioctl(c_fd,SIOC_WANPIPE_SOCK_STATE,0);
					mprintf("audio_thread: Sock state is %s\n",
						(err == 0) ? "CONNECTED" :
						(err == 1) ? "DISCONNECTED" :
						"CONNECTING");
					if (err == 1){
						mprintf("audio_thread: Terminating due to disconnection\n");
						connected =0;
					}
				}
			}
/*			if (FD_ISSET(c_fd2,&oob_fds)){
				nbytes = recv(c_fd, buf, sizeof(buf), MSG_OOB);
				if(nbytes < 0 ) {
					mprintf("audio_thread: Failed to receive OOB \n");
					err = ioctl(c_fd,SIOC_WANPIPE_SOCK_STATE,0);
					mprintf("audio_thread: Sock state is %s\n",
						(err == 0) ? "CONNECTED" :
						(err == 1) ? "DISCONNECTED" :
						"CONNECTING");
					if (err == 1){
						mprintf("audio_thread: Terminating due to disconnection\n");
						connected =0;
					}
				}
			} */
			if (FD_ISSET(c_fd,&port_fds)){
				if ((nbytes = recv(c_fd, buf, sizeof(buf), 0)) <= 0) {
					if (nbytes == 0) {
						mprintf("[%s][%d] Socket hangup ...\n",
							c->iface->card, c->iface->name);
					} else {
						mprintf("[%s][%d] Socket error[%d] ...\n",
							c->iface->card, c->iface->name,errno);
					}
				} else {
					if ((nbytes = recv(c_fd2, buf2, sizeof(buf2), 0)) <= 0) {
						if (nbytes == 0) {
							mprintf("[%s][%d] Socket hangup ...\n",
								c->iface->card, c->iface->name);
						} else {
							mprintf("[%s][%d] Socket error[%d] ...\n",
								c->iface->card, c->iface->name,errno);
						}
					} else {
						for(i=0;i<(nbytes-16);i++){
							buf[16+i] = bit_inv[buf[16+i]];
							buf2[16+i] = bit_inv[buf2[16+i]];
						}
						prepare_rec_buf_log(&buf[16], &buf2[16],
								    c->v, c->chan_num,
								    encoding);
					}
				}
			} /*else if (FD_ISSET(c_fd2,&port_fds)){
				// Get data from socket - record 
				if ((nbytes = recv(c_fd2, buf2, sizeof(buf2), 0)) <= 0) {
					if (nbytes == 0) {
						mprintf("[%s][%d] Socket hangup ...\n",
							c->iface->card, c->iface->name);
					} else {
						mprintf("[%s][%d] Socket error[%d] ...\n",
							c->iface->card, c->iface->name,errno);
					}
				} else {
					if ((nbytes = recv(c_fd, buf, sizeof(buf), 0)) <= 0) {
						if (nbytes == 0) {
							mprintf("[%s][%d] Socket hangup ...\n",
								c->iface->card, c->iface->name);
						} else {
							mprintf("[%s][%d] Socket error[%d] ...\n",
								c->iface->card, c->iface->name,errno);
						}
					} else {
						for(i=0;i<(nbytes-16);i++){
							buf[16+i] = bit_inv[buf[16+i]];
							buf2[16+i] = bit_inv[buf2[16+i]];
						}
						prepare_rec_buf_log(&buf[16], &buf2[16],
								    c->v, c->chan_num,
								    encoding);
					}
				}
			} */
			nbytes=0;
		}
	}
	c->v->rxdf[c->chan_num]->SetState(0);
	c->v->txdf[c->chan_num]->SetState(0);
	close(c_fd);
	close(c_fd2);
	gettimeofday(&logtime,NULL);
	mprintf("[%ld:%ld]Finished LOGGING audio on channel [%d] interface [%s] encoding[%s]\n",
		logtime.tv_sec,logtime.tv_usec,c->chan_num,c->iface->name,
		(VPB_ALAW==encoding?"alaw":(VPB_MULAW==encoding?"mulaw":"raw")));
	pthread_exit((void*)0);
} //}}}
#endif

int check_encoding(pri_chan * c)
{ //{{{
	int encoding;

	mprintf("check_encoding: Checking channel %d [layer1: %s (%d)]...\n",
		c->chan_num,
		(c->layer1 == PRI_LAYER_1_ALAW ? "alaw"
		    : (c->layer1 == PRI_LAYER_1_ULAW ? "mulaw" : "unknown")),
		c->layer1);
	// check encoding
	if( (c->layer1 == PRI_LAYER_1_ULAW && c->encoding != VPB_MULAW)  // Compare if mulaw
	 || (c->layer1 == PRI_LAYER_1_ALAW && c->encoding != VPB_ALAW) ) // Compare if alaw
	{
		//mprintf("Fixing Layer1(0x%02X) and encoding(%d)!\n",c->layer1,c->encoding);
		// Encoding doesnt match layer1 from call setup, fixit!
		if (c->layer1 == PRI_LAYER_1_ULAW){
			encoding = VPB_MULAW;
		} else if (c->layer1 == PRI_LAYER_1_ALAW){
			encoding = VPB_ALAW;
		} else {
			encoding = VPB_RAW;
		}
	} else {
		encoding = c->encoding;
	}
	if (c->transcap == PRI_TRANS_CAP_DIGITAL){
		encoding = VPB_RAW;
	}
	mprintf("check_encoding: channel encoding set to [%s].\n",
		(VPB_ALAW==encoding ? "alaw"
				    :(VPB_MULAW==encoding ? "mulaw" : "raw")));
	return encoding;
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: audio_thread_multi
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04
	
	Connects to the correct wanpipe interface and gets/puts the audio onto
	the correct driver FIFO's for multiple channels on single interface

\*--------------------------------------------------------------------------*/

static void *audio_thread_multi(void *pv)
{ //{{{
	unsigned char bit_inv[256];
	AUDIO_THREAD_MULTI * info;
	info = (AUDIO_THREAD_MULTI *)pv;
	pri_chan * c;
	VPBREG * v = info->v;
	int c_fd,i,x,tmp;
	int ret,nbytes,err;
	unsigned char * buf;
	unsigned char chbuf[32][NBUF];
	int chan_state[33]={0};
	int encoding[33]={0};
	int channels=0;

	// Select stuff
	fd_set port_fds,write_fds,oob_fds;
	struct timeval tv;

	// Work out number of channels
	channels = info->end_channel - info->start_channel +1;

	// Get buffer memory
	buf = new unsigned char[16+(channels*NBUF)];

	pthread_detach(pthread_self());
	for(i=0;i<256;i++){
		bit_inv[i] = (i&1?0x80:0);
		bit_inv[i] |= (i&2?0x40:0);
		bit_inv[i] |= (i&4?0x20:0);
		bit_inv[i] |= (i&8?0x10:0);
		bit_inv[i] |= (i&0x10?0x08:0);
		bit_inv[i] |= (i&0x20?0x04:0);
		bit_inv[i] |= (i&0x40?0x02:0);
		bit_inv[i] |= (i&0x80?0x01:0);
	}

	mprintf("Starting audio thread for channels [%d] to [%d], total [%d]"
		" using card [%s] interface [%s]\n",
		info->start_channel, info->end_channel,
		channels, info->card, info->interface);
	info->run = 1;
	while( info->run ){
            c_fd = wanconnect((char *)info->card,(char *)info->interface);
            if (c_fd < 0) {
                mprintf("Audio thread failed to connect interface [%s] on card '%s'\n",
                        info->interface, info->card);
                goto done;
            }

            mprintf("Audio thread connected to interface[%s] on card [%s]"
                    " using socket on fd[%d]\n", info->interface, info->card, c_fd);
            while( info->run ){
		/* Set up file descriptor set for select()ing on sockets */
		FD_ZERO(&port_fds);
		FD_ZERO(&write_fds);
		FD_ZERO(&oob_fds);
		FD_SET(c_fd,&port_fds);
		FD_SET(c_fd,&write_fds);
		FD_SET(c_fd,&oob_fds);

		/* wait for a socket to be ready */
		tv.tv_sec = 1;
		tv.tv_usec = 1000;
		ret = select(c_fd+1,&port_fds,NULL,&oob_fds,&tv);
		if( ret <= 0 ) continue;

		if (FD_ISSET(c_fd,&oob_fds)){
			nbytes = recv(c_fd, buf, channels * NBUF + 16, MSG_OOB);
			if(nbytes < 0 ) {
				mprintf("audio_thread_multi: Failed to receive OOB \n");
				err = ioctl(c_fd,SIOC_WANPIPE_SOCK_STATE,0);
				mprintf("audio_thread_multi: Sock state is %s\n",
					(err == 0) ? "CONNECTED" :
					(err == 1) ? "DISCONNECTED" :
					"CONNECTING");
				if (err == 1){
					mprintf("audio_thread_multi: disconnected\n");
					goto disconnected;
				}
			}
		}
		if (FD_ISSET(c_fd,&port_fds)){
			/* Get data from socket - record */
			nbytes = recv(c_fd, buf, channels * NBUF + 16, 0);
			// got error or connection closed by client
			if (nbytes == 0) {
				// connection closed
				mprintf("audio_thread_multi: Socket hangup on [%s] [%s]\n",
					info->card, info->interface);
			} else if (nbytes < 0) {
				mprintf("audio_thread_multi: Socket error[%d] on [%s] [%s]\n",
					errno, info->card, info->interface);
				if( errno != EINTR && errno != EAGAIN )
					goto disconnected;
			} else {
				for(i=0; i < (nbytes-16); ++i){
					//buf[16+i] = bit_inv[buf[16+i]];
					chbuf[i%channels][i/channels] = bit_inv[buf[16+i]];
				}
				/* Ok, split audio up into channels */
				for (i=info->start_channel,x=0; i <= info->end_channel; ++i, ++x){
					c = &((pri_chan *)v->chans)[i];
					tmp = c->run;
					if (tmp == 1){
						if (chan_state[i] != 1){
							mprintf("audio_thread_multi: Starting audio on [%d](%d/%d)\n",
								c->chan_num,i,x);
							chan_state[i]=1;
							c->v->rxdf[c->chan_num]->SetState(1);
							c->v->txdf[c->chan_num]->SetState(1);
							c->v->rxdf[c->chan_num]->Flush();
							c->v->txdf[c->chan_num]->Flush();
							encoding[i]=check_encoding(c);
						}
						/* Send data up FIFO - record  */
						prepare_rec_buf(chbuf[x], c->v,
								c->chan_num,encoding[i]);
						/* Get data from FIFO - play */
						prepare_play_buf(chbuf[x], c->v,
								 c->chan_num,encoding[i]);
					} else {
						if (chan_state[i] == 1){
							mprintf("audio_thread_multi: Finishing audio on [%d]\n",
								c->chan_num);
							chan_state[i]=0;
							c->v->rxdf[c->chan_num]->SetState(0);
							c->v->txdf[c->chan_num]->SetState(0);
						}
					}
				}

				for(i=0; i < (nbytes-16); ++i){
					buf[16+i] = bit_inv[chbuf[i%channels][i/channels]];
				}
				/* Send FIFO up socket  - play */
				err = send(c_fd,buf,sizeof(unsigned char [16+(channels*NBUF)]),0);
			}
		}
		nbytes=0;
	    }

          disconnected:
	    close(c_fd);
	}
	mprintf("Finished audio thread for channels [%d] to [%d] using card [%s] interface [%s]\n",
		info->start_channel,info->end_channel,info->card,info->interface);

    done:
	delete [] buf;
	pthread_exit((void*)0);
} //}}}

/*--------------------------------------------------------------------------*\

	FUNCTION.: audio_thread_log_multi
	AUTHOR...: Ben Kramer
	DATE.....: 19/07/05
	
	Connects to the correct wanpipe interfaces and gets/puts the audio onto
	the correct driver FIFO's after combing them from both interfaces!

\*--------------------------------------------------------------------------*/

static void *audio_thread_log_multi(void *pv)
{ //{{{
	AUDIO_THREAD_MULTI * info;
	info = (AUDIO_THREAD_MULTI *)pv;
	pri_chan * c;
	VPBREG * v = info->v;

	unsigned char bit_inv[256];
	int i,x;
	int c_fd,c_fd2,selfd;
	int ret, err;
	int nbytes, nbytes2;
	unsigned char * buf;
	unsigned char * buf2;
	unsigned char chbuf[32][NBUF];
	unsigned char chbuf2[32][NBUF];
	char card2[32];
	char interface2[32];
	int chan_state[33]={0};
	int encoding[33]={0};
	int channels=0;
	char *tmpstr;

//	api_rx_element_t* api_rx_el;

	// Select stuff
	fd_set port_fds,write_fds,oob_fds;
	struct timeval tv;

        // Work out number of channels
        channels = info->end_channel - info->start_channel +1;

        // Get buffer memory
        buf = new unsigned char[16+(channels*NBUF)];
        buf2 = new unsigned char[16+(channels*NBUF)];

	// Split up channel and card names
	tmpstr = strchr((char *)info->interface,',');
	if (tmpstr != NULL){
		*tmpstr = '\0';
		++tmpstr;
		strcpy(interface2,tmpstr);
	}
	tmpstr = strchr((char *)info->card,',');
	if (tmpstr != NULL){
		*tmpstr = '\0';
		++tmpstr;
		strcpy(card2,tmpstr);
	}

	pthread_detach(pthread_self());
	for(i=0;i<256;i++){
		bit_inv[i] = (i&1?0x80:0);
		bit_inv[i] |= (i&2?0x40:0);
		bit_inv[i] |= (i&4?0x20:0);
		bit_inv[i] |= (i&8?0x10:0);
		bit_inv[i] |= (i&0x10?0x08:0);
		bit_inv[i] |= (i&0x20?0x04:0);
		bit_inv[i] |= (i&0x40?0x02:0);
		bit_inv[i] |= (i&0x80?0x01:0);
	}

	mprintf("Starting LOGGING audio thread for channels [%d] to [%d],"
		" total [%d] using card [%s] interface [%s]\n",
		info->start_channel, info->end_channel, channels,
		info->card, info->interface);
	info->run = 1;
	while( info->run ){
	    c_fd = wanconnect((char *)info->card,(char *)info->interface);
	    c_fd2 = wanconnect(card2,interface2);
            if (c_fd < 0) {
		mprintf("Failed to connect to channel interface on card '%s' iface '%s': %d\n",
			info->card,info->interface, c_fd);
		goto done;
            } else if (c_fd2 < 0) {
		mprintf("Failed to connect to channel interface on card '%s' iface '%s': %d\n",
			card2, interface2, c_fd2);
		goto done;
            }

            mprintf("Connected to interfaces [%s] and [%s]\n",info->interface,interface2);
            selfd = (c_fd2 > c_fd) ? c_fd2 : c_fd;
	    while( info->run ){
		/* Set up file descriptor set for select()ing on sockets */
		FD_ZERO(&port_fds);
		FD_ZERO(&write_fds);
		FD_ZERO(&oob_fds);
		FD_SET(c_fd,&port_fds);
		FD_SET(c_fd,&write_fds);
		FD_SET(c_fd,&oob_fds);

		/* wait for a socket to be ready */
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		ret = select(c_fd+1,&port_fds,NULL,&oob_fds,&tv);
		if( ret <= 0 ) continue;

		if (FD_ISSET(c_fd,&oob_fds)){
			nbytes = recv(c_fd, buf, channels * NBUF + 16, MSG_OOB);
			if(nbytes < 0 ) {
				mprintf("audio_thread: Failed to receive OOB \n");
				err = ioctl(c_fd,SIOC_WANPIPE_SOCK_STATE,0);
				mprintf("audio_thread: Sock state is %s\n",
					(err == 0) ? "CONNECTED" :
					(err == 1) ? "DISCONNECTED" :
					"CONNECTING");
				if (err == 1){
					mprintf("audio_thread: disconnected\n");
					goto disconnected;
				}
			}
		}
		if (FD_ISSET(c_fd,&port_fds)){
			nbytes = recv(c_fd, buf, channels * NBUF + 16, 0);
			if( nbytes == 0 ) {
				mprintf("[%s][%d] Socket hangup ...\n",
					info->card, info->interface);
				goto disconnected;
			} else if( nbytes < 0 ) {
				mprintf("[%s][%d] Socket error[%d] ...\n",
					info->card, info->interface,errno);
				if( errno != EINTR && errno != EAGAIN )
					goto disconnected;
			} else {
				nbytes2 = recv(c_fd2, buf2, channels * NBUF + 16, 0);
				if( nbytes2 == 0 ) {
					mprintf("[%s][%d] Socket hangup ...\n",
						card2, interface2);
					goto disconnected;
				} else if( nbytes2 < 0 ) {
					mprintf("[%s][%d] Socket error[%d] ...\n",
						card2, interface2,errno);
					if( errno != EINTR && errno != EAGAIN )
						goto disconnected;
				} else {
					for(i=0;i<(nbytes-16);i++){
						chbuf[i%channels][i/channels] = bit_inv[buf[16+i]];
						chbuf2[i%channels][i/channels] = bit_inv[buf2[16+i]];
					}
					for (i=info->start_channel,x=0; i <= info->end_channel; ++i,++x){
						c = &((pri_chan *)v->chans)[i];
						if( c->run ){
							if (chan_state[i] != 1){
								mprintf("audio_thread_multi: Starting audio on [%d](%d/%d)\n",
									c->chan_num,i,x);
								chan_state[i]=1;
								c->v->rxdf[c->chan_num]->SetState(1);
								c->v->txdf[c->chan_num]->SetState(1);
								c->v->rxdf[c->chan_num]->Flush();
								c->v->txdf[c->chan_num]->Flush();
								encoding[i]=check_encoding(c);
							}
							/* Send data up FIFO - record  */
							prepare_rec_buf_log(chbuf[x],chbuf2[x],
									    c->v,
									    c->chan_num,encoding[i]);
						} else {
							if (chan_state[i] == 1){
								mprintf("audio_thread_multi: Finishing audio on [%d]\n",
									c->chan_num);
								chan_state[i]=0;
								c->v->rxdf[c->chan_num]->SetState(0);
								c->v->txdf[c->chan_num]->SetState(0);
							}
						}
					}
				}
			}
		}
		nbytes=0;
	    }

          disconnected:
	    close(c_fd);
	    close(c_fd2);
	}
	mprintf("Finished LOGGING audio thread for channels [%d] to [%d]"
		" using card [%s] interface [%s]\n",
		info->start_channel,info->end_channel,info->card,info->interface);

    done:
	delete [] buf;
	delete [] buf2;
	pthread_exit((void*)0);
} //}}}


/*--------------------------------------------------------------------------*\

	FUNCTION.: hw_gain_vector
	AUTHOR...: Ben Kramer
	DATE.....: xx/11/04
	
	This emulates the hardware gain as per the other cards

\*--------------------------------------------------------------------------*/
int hw_gain_vector(float g, short *v, int n)
{
        int i;
        float tmp;

	if( fabs(g) < std::numeric_limits<float>::epsilon() )
		return(0);
	float l = pow(10.0,(g)/20.0);
//	mprintf("Applying [%f/%f] gain\n",g,l);
        for ( i = 0; i<n; i++) {
                tmp = l*v[i];
                if (tmp > 32767.0) {
                        tmp = 32767.0;
//			mprintf("Clipping audio!\n");
		}
                if (tmp < -32768.0) {
                        tmp = -32768.0;
//			mprintf("Clipping audio!\n");
		}
                v[i] = (short)tmp;
        }
        return(i);
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: make_isdn_call
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Actually places a call out of the system.

\*--------------------------------------------------------------------------*/
static int make_isdn_call(word * d, VPBREG * v)
{
	word * mp = (word *)d;
	USHORT ch = 0;
	int handle = -1;
	char number[21];
	char cid[21];
	int ret;
	int ulayer1=PRI_LAYER_1_ALAW;
	int layer1;
	int dialplan;
	int localdialplan;
	char *calledsubaddr;
	int i;
	// Set up dial plans
	OP_CARD *cardinfo = (OP_CARD *)v->cardinfo;
	dialplan = cardinfo->dialplan;
	localdialplan = cardinfo->localdialplan;

	memcpy(number,&mp[4],sizeof(char)*21);
	memcpy(cid,&mp[15],sizeof(char)*21);
	ch = (USHORT)mp[2];
	layer1 = (int)mp[3];
	handle=mapdevtohndle(v->cardnum,ch);
	int idx;

	try {
		idx = get_idx_from_ch(v,ch);
	} catch ( const Wobbly &w ) {
		mprintf("[%d/??] make_isdn_call: Cant work out IDX from ch[%d]!\n",
			v->cardnum,ch);
		return -1;
	}

	pri_chan *chans = (pri_chan *)v->chans;
	OpenPri* pThis = (OpenPri *)v->hostdsp;
	struct pri *pri = pThis->pri;

	if (!pri){
		mprintf("[%d/%d] No Primary channel [%d]!\n",v->cardnum,chans[idx].chan_num,pri);
		return 1;
	}

	// Check that the channel is not already in a call
	if (chans[idx].state != CH_IDLE) {
		mprintf("[%d/%d] Channel is not idle! is [%d]\n",
			v->cardnum,chans[idx].chan_num,chans[idx].state);
		return 2;
	}
	// Create a new call
	if (!(chans[idx].call = pri_new_call(pri))){
		mprintf("[%d/%02d] Couldnt create a new call!\n",v->cardnum,chans[idx].chan_num);
		return 3;
	}
	chans[idx].state = CH_SETUP;
	if (layer1 >0){
		ulayer1=layer1;
		mprintf("[%d/%02d] Using passed Encoding (%d)",v->cardnum,chans[idx].chan_num,ulayer1);
	} else if (chans[idx].encoding == VPB_ALAW){
		ulayer1 = PRI_LAYER_1_ALAW;
		mprintf("[%d/%02d] Using ALAW in new call",v->cardnum,chans[idx].chan_num);
	} else if (chans[idx].encoding == VPB_MULAW){
		ulayer1 = PRI_LAYER_1_ULAW;
		mprintf("[%d/%02d] Using MULAW in new call",v->cardnum,chans[idx].chan_num);
	} else {
		ulayer1 = PRI_LAYER_1_ALAW;
		mprintf("[%d/%02d] Defaulting to ALAW in new call",v->cardnum,chans[idx].chan_num);
	}

	// Check number for sub-address and copy into local variables
	calledsubaddr=NULL;
	i=0;
	while(number[i] != '\0'){
		if(number[i] == '*'){
			calledsubaddr = &number[i+1];
			number[i]='\0';
			mprintf("[%d/%02d] Called SubAddress [%s]\n",
				v->cardnum,chans[idx].chan_num, calledsubaddr);
			break;
		}
		i++;
	}

	ret = pri_call(	pri,
			chans[idx].call,
			PRI_TRANS_CAP_SPEECH,
			chans[idx].iface->channel,
			1,
			0,
			cid,
			localdialplan,
			cid,
			PRES_ALLOWED_USER_NUMBER_PASSED_SCREEN,
			number,
			dialplan,
			calledsubaddr,
			ulayer1,
			NULL);
	if (ret ==1 ){
		mprintf("[%d/%02d] Couldnt set up call!",v->cardnum,chans[idx].chan_num);
		return 4;
	}
	return 0;
}


/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::OpenPri
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Constructor calls priHostDSPOpen()

\*--------------------------------------------------------------------------*/
OpenPri::OpenPri(VPBREG *ivr)
    : vr( ivr )
    , active_threads( 0 )
    , pri( NULL )
    , pri2( NULL )
{
	priHostDSPOpen();
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::~OpenPri
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Destructore calls priHostDSPClose()

\*--------------------------------------------------------------------------*/
OpenPri::~OpenPri()
{
	priHostDSPClose();
	free( pri );
	free( pri2 );
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::chan_state()
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Returns the state of a channel

\*--------------------------------------------------------------------------*/
int OpenPri::chan_state(int ch)
{
	try {
		pri_chan *chans = (pri_chan *)vr->chans;
		return chans[get_idx_from_ch(vr,ch)].state;
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_state: Cant work out IDX from ch[%d]!\n",ch);
		throw;	// Nothing much else we can do here.
	}
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::chan_layer1()
	AUTHOR...: Ben Kramer
	DATE.....: xx/07/05
	
	Returns the layer1 of a channel

\*--------------------------------------------------------------------------*/
int OpenPri::chan_layer1(int ch)
{
	try {
		pri_chan *chans = (pri_chan *)vr->chans;
		return chans[get_idx_from_ch(vr,ch)].layer1;
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_layer1: Cant work out IDX from ch[%d]!\n",ch);
		throw;	// Nothing much else we can do here.
	}
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::chan_transcap()
	AUTHOR...: Ben Kramer
	DATE.....: 11/07/05
	
	Returns the transmit capability of a channel

\*--------------------------------------------------------------------------*/
int OpenPri::chan_transcap(int ch)
{
	try {
		pri_chan *chans = (pri_chan *)vr->chans;
		return chans[get_idx_from_ch(vr,ch)].transcap;
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_transcap: Cant work out IDX from ch[%d]!\n",ch);
		throw;	// Nothing much else we can do here.
	}
}


/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::chan_cinfo()
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Provides the call info of a channel

\*--------------------------------------------------------------------------*/
int OpenPri::chan_cinfo(int ch, void *info)
{
	try {
		int idx = get_idx_from_ch(vr,ch);
		pri_chan *chans = (pri_chan *)vr->chans;
		if( info != NULL && chans[idx].cid != NULL ){
			memcpy(info,chans[idx].cid,sizeof(isdncid));
			return 0;
		}
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_cinfo: Cant work out IDX from ch[%d]!\n",ch);
	}
	return -1;
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::chan_cause()
	AUTHOR...: Ben Kramer
	DATE.....: xx/12/04
	
	Provides the cause code from the last call

\*--------------------------------------------------------------------------*/
int OpenPri::chan_cause(int ch)
{
	try {
		pri_chan *chans = (pri_chan *)vr->chans;
		return chans[get_idx_from_ch(vr,ch)].cause;
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_cause: Cant work out IDX from ch[%d]!\n",ch);
		throw;	// Nothing much else we can do here.
	}
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::call_proceeding()
	AUTHOR...: Ben Kramer
	DATE.....: 11/07/05
	
	Sends a call proceeding

\*--------------------------------------------------------------------------*/
int OpenPri::call_proceeding(int ch)
{
	pri_chan *chans = (pri_chan *)vr->chans;
	int idx;

	try {
		idx = get_idx_from_ch(vr,ch);
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_proceeding: Cant work out IDX from ch[%d]!\n",ch);
		return 1;
	}
	if( chans[idx].state != CH_RING ) {
		mprintf("[%d/%d]Cant send call proceeding/alerting",
			vr->cardnum,chans[idx].chan_num);
		mprintf(" not in ring state (state[%d])\n",chans[idx].state);
		return 1;
	}

	mprintf("[%d/%d] Call proceeding/alerting \n",
		vr->cardnum,chans[idx].chan_num);

	/* send pri message */
	#if 0
	/* Changed to pri_acknowldege aka PROCEEDING+ALERTING */
	//ret = pri_need_more_info(pri,chans[idx].call,chans[idx].iface->channel,1);
	ret = pri_acknowledge(pri,chans[idx].call,chans[idx].iface->channel,0);
	//ret = pri_progress(pri,chans[idx].call,chans[idx].iface->channel,8);
	//ret = pri_proceeding(pri,chans[idx].call,chans[idx].iface->channel,0);
	if (ret == 0) {
		chans[idx].state = CH_RING;
		return 0;
	} else {
		mprintf("[%d/%d] Failed to send call proceeding!\n",
			vr->cardnum,chans[idx].chan_num);
		return 1;
	}
	#endif

	if( pri_acknowledge(pri,chans[idx].call,chans[idx].iface->channel,1) != 0 ) {
		mprintf("[%d/%d] Failed to send call proceeding/alerting!\n",
			vr->cardnum,chans[idx].chan_num);
		return 1;
	}
	chans[idx].state = CH_RING;
	return 0;
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: OpenPri::make_isdn_call_api
	AUTHOR...: Ben Kramer
	DATE.....: 11/07/05
	
	Actually places a call out of the system.

\*--------------------------------------------------------------------------*/
int OpenPri::make_isdn_call_api(int ch, char *number, char *cid, int layer1,
				int transcap, unsigned char *lowlayercompatibility)
{
	int ret;
	int idx;
	int ulayer1=PRI_LAYER_1_ALAW;
	char *calledsubaddr;
	int i;
	// Set up dial plans
	OP_CARD *cardinfo = (OP_CARD *)vr->cardinfo;
	int dialplan = cardinfo->dialplan;
	int localdialplan = cardinfo->localdialplan;
	pri_chan *chans = (pri_chan *)vr->chans;

	try {
		idx = get_idx_from_ch(vr,ch);
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::make_isdn_call_api: Cant work out IDX from ch[%d]!\n",ch);
		throw;
	}

	if (!pri){
		mprintf("[%d/%d] No Primary channel [%d]!\n",
			vr->cardnum,chans[idx].chan_num,pri);
		return 1;
	}

	// Check that the channel is not already in a call
	if (chans[idx].state != CH_IDLE) {
		mprintf("[%d/%d] Channel is not idle! is [%d]\n",
			vr->cardnum,chans[idx].chan_num,chans[idx].state);
		return 2;
	}
	// Create a new call
	if (!(chans[idx].call = pri_new_call(pri))){
		mprintf("[%d/%02d] Couldnt create a new call!\n",
			vr->cardnum,chans[idx].chan_num);
		return 3;
	}
	chans[idx].state = CH_SETUP;
	if (layer1 > 0){
		ulayer1=layer1;
		mprintf("[%d/%02d] Using passed Encoding (%d)\n",
			vr->cardnum,chans[idx].chan_num,ulayer1);
		chans[idx].layer1 = ulayer1;
	} else if (chans[idx].encoding == VPB_ALAW){
		ulayer1 = PRI_LAYER_1_ALAW;
		mprintf("[%d/%02d] Using ALAW in new call\n",
			vr->cardnum,chans[idx].chan_num);
	} else if (chans[idx].encoding == VPB_MULAW){
		ulayer1 = PRI_LAYER_1_ULAW;
		mprintf("[%d/%02d] Using MULAW in new call\n",
			vr->cardnum,chans[idx].chan_num);
	} else {
		ulayer1 = PRI_LAYER_1_ALAW;
		mprintf("[%d/%02d] Defaulting to ALAW in new call\n",
			vr->cardnum,chans[idx].chan_num);
	}
	// Check for trans cap
	if(transcap < 1){
		transcap = PRI_TRANS_CAP_SPEECH;
	} else if( transcap == PRI_TRANS_CAP_DIGITAL
		|| transcap == PRI_TRANS_CAP_RESTRICTED_DIGITAL ){
		chans[idx].encoding = VPB_RAW;
		ulayer1 = -1;
	}
	chans[idx].transcap = transcap;

	mprintf("[%d/%02d] Checking for Called SubAddress in [%s]\n",
		vr->cardnum,chans[idx].chan_num, number);
	// Check number for sub-address and copy into local variables
	calledsubaddr=NULL;
	i=0;
	while(number[i] != '\0'){
		if(number[i] == '*'){
			calledsubaddr = &number[i+1];
			number[i]='\0';
			mprintf("[%d/%02d] Called SubAddress [%s]\n",
				vr->cardnum,chans[idx].chan_num, calledsubaddr);
			break;
		}
		i++;
	}
	mprintf("[%d/%02d] Done Checking for Called SubAddress in [%s]\n",
		vr->cardnum,chans[idx].chan_num, number);

	ret = pri_call(	pri,
			chans[idx].call,
			transcap,
			chans[idx].iface->channel,
			1,
			0,
			cid,
			localdialplan,
			cid,
			PRES_ALLOWED_USER_NUMBER_PASSED_SCREEN,
			number,
			dialplan,
			calledsubaddr,
			ulayer1,
			lowlayercompatibility);
	if (ret ==1 ){
		mprintf("[%d/%02d] Couldnt set up call!",vr->cardnum,chans[idx].chan_num);
		return 4;
	}
	return 0;
}

int OpenPri::chan_lowlayercompatibility(int ch, unsigned char *lowlayercompatibility)
{
	int idx;

	try {
		idx = get_idx_from_ch(vr,ch);
	} catch ( const Wobbly &w ) {
		mprintf("OpenPri::chan_lowlayercompatibility: Cant work out IDX from ch[%d]!\n",ch);
		return -1;
	}
	pri_chan *chans = (pri_chan *)vr->chans;
	for(int i=0; i < chans[idx].lowlayercompatibility[0] + 1; ++i){
		lowlayercompatibility[i]=chans[idx].lowlayercompatibility[i];
		mprintf("LLC (%d) (%02X)",i,lowlayercompatibility[i]);
	}
	return VPB_OK;
}

#if 0
static int slot2chan(int slotmap, int len)
{
	int x;
	for (x=1;x<=len;x++){
		if(slotmap & (1 << (len - x)))
			return(x);
	}
	return 0;
}
#endif

static int findfreechan(VPBREG *v)
{
	pri_chan *chans;
	int	num_chans,i;
	chans = (pri_chan *)v->chans;
	num_chans = v->numch;
	for (i=0;i<num_chans;i++){
		if (chans[i].state == CH_IDLE){
			return(i);
		}
	}
	return(-1);
}
