/*---------------------------------------------------------------------------*\

    FILE....: VPBREG.CPP
    TYPE....: C++ Functions
    AUTHOR..: David Rowe
    DATE....: 12/9/99

    Helper functions to set up the configuration database known as the DSP 
    registry.

\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

         Voicetronix Voice Processing Board (VPB) Software

         Copyright (C) 1999-2001 Voicetronix www.voicetronix.com.au

         This library is free software; you can redistribute it and/or
         modify it under the terms of the GNU Lesser General Public
         License as published by the Free Software Foundation; either
         version 2.1 of the License, or (at your option) any later version.

         This library is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty of
         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
         License along with this library; if not, write to the Free Software
         Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	 USA

\*---------------------------------------------------------------------------*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "vpbreg.h"
#include "wobbly.h"
#include "mess.h"
#include "vpbapi.h"
#include "apifunc.h"
#include "generic.h"
#include "dspfifo.h"
#include "verbose.h"
#include "vpbdial.h"
#include "mapdev.h"
#include "llconf.h"

// Compile-time constants --------------------------------------------

#define	NUM_CARDS		1     // def number of VPBs in this PC
#define	BASE_ADDRESS	        0x310 // def base address of VPB

#define	NUM_CHANNELS_VPB4       4     // number of channels in VPB4, V4PCI
#define	NUM_CHANNELS_VPB8L      8     // number of channels in VPB8L

#define	SIZE_MESS_Q_HDSP	4*256   // size of message FIFOs (16-bit words) for Host DSP's
#define	SIZE_MESS_Q_OL4		256   // size of message FIFOs (16-bit words) for OpenLine4 DSP

// DR 24/9/02 - note changing any of these 3 below may kill host based EC
// code in dspfifo.cpp may also need mods

#define	SIZE_RX_Q		800   // size of Rx (up) signal FIFOs
#define	SIZE_TX_Q		800   // size of Tx (down) signal FIFOs
#define	LENGTH_FRAME	        160   // samples in each processing frame

// device driver types

#define	DD_ISA_PORT		0     // ISA port I/O only device driver
#define DD_ISA_RELAY            1     // ISA relay FIFO device driver
#define DD_PCI                  2     // PCI driver
#define DD_V12PCI               3     // V12PCI driver
#define DD_V6PCI                3     // V6PCI driver
#define DD_PRI                  4     // OpenPri driver
#define DD_VTCORE               5     // VTCore card driver

#define	SIZE_RELAY_BUF	16000	      // size of relay buffer in device driver

// default gains (VPB4 & V4PCI & V12PCI)

#define DEF_RECORD_GAIN_ISA      0.0
#define DEF_PLAY_GAIN_ISA        0.0
#define DEF_RECORD_GAIN_PCI      6.0
#define DEF_PLAY_GAIN_PCI        7.0
#define DEF_RECORD_GAIN_V12PCI   3.0
#define DEF_PLAY_GAIN_V12PCI     12.0
#define DEF_RECORD_GAIN_SV12PCI   -6.0
#define DEF_PLAY_GAIN_SV12PCI     7.0
#define DEF_RECORD_GAIN_PRI      1.0
#define DEF_PLAY_GAIN_PRI        1.0

// default codec balance registers

#define DEF_BAL1                 0xc7
#define DEF_BAL2                 0x00
#define DEF_BAL3                 0x00

// path and name of firmware file

#define	FIRMWARE_FILE_ISA	"/etc/vpb/vpbmain_isa.out"
#define	FIRMWARE_FILE_ISA_VPB8L	"/etc/vpb/vlcmain.out"
#define	FIRMWARE_FILE_PCI	"/etc/vpb/vpbmain_pci.out"

/*-------------------------------------------------------------------------*\

			     STATICS

\*-------------------------------------------------------------------------*/

static int vpb_num_cards = NUM_CARDS;
static int def_base[1] = {BASE_ADDRESS};
static int *vpb_base = def_base;
static char *vpb_firmware_file = FIRMWARE_FILE_PCI;
static int vpb_model = VPB_V4PCI;
static int defaultsused = 1;

static int def_bal1 = DEF_BAL1;
static int def_bal2 = DEF_BAL2;
static int def_bal3 = DEF_BAL3;

static USHORT def_hp = 0;
static float  defPlayGain = DEF_PLAY_GAIN_PCI;
static float  defRecordGain = DEF_RECORD_GAIN_PCI;
static float  defSPlayGain = DEF_PLAY_GAIN_SV12PCI;
static float  defSRecordGain = DEF_RECORD_GAIN_SV12PCI;
static float  defRecordGainDown = 20.0;	// DTMF CutThrough control <20 disable

static USHORT def_hostec = 0;
static int def_echo_en = 0;
static USHORT def_dighyb = 0;
static char *def_dighybfile = NULL;

/*-------------------------------------------------------------------------*\

			      FUNCTIONS

\*-------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*\

	FUNCTION.: vpb_config
	AUTHOR...: David Rowe
	DATE.....: 24/8/00

	Set installation specific config info, call before vpb_open().  If
	this function is called, it overrides any other settings (for example
	compile-time defaults or env vars).

\*--------------------------------------------------------------------------*/

int WINAPI vpb_config(int num_cards, int *bases, char *firmware_file, int model) {

	assert((num_cards > 0) && (num_cards <MAX_VPB));
	assert(bases != NULL);
	assert(firmware_file != NULL);

	vpb_num_cards = num_cards;
	vpb_base = bases;
	vpb_firmware_file = firmware_file;
	vpb_model = model;

	defaultsused = 0;

	return VPB_OK;
}

/*-------------------------------------------------------------------------*\

			      CLASS

\*-------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*\

	FUNCTION.: ~VPBRegister
	AUTHOR...: John Kostogiannis
	DATE.....: 18/11/97

	Closes the VPB registry.

\*--------------------------------------------------------------------------*/

VPBRegister::~VPBRegister() {
	// free allocated VPB registry memory 

	conf_release_list( config_data );
	delete ci;     ci = NULL;
	free(reg);     reg = NULL;
	free(handles); handles = NULL;
}

/*--------------------------------------------------------------------------*\

	FUNCTION.: VPBRegister
	AUTHOR...: David Rowe
	DATE.....: 12/9/99

	Creates the VPB registry using compile-time constants.

\*--------------------------------------------------------------------------*/

VPBRegister::VPBRegister(USHORT *num)
//      USHORT *num;    number of VPBs
{
	*num = vpb_num_cards;
        int    hybdefs_overwrite;
	int	n_pg=0, n_rg=0;
	int	loc_model=-1;
	int	*cards= NULL;
	//int	numOL=0,numOS=0,numOP=0;

	next_handle=0;

	VPBREG *preg; // ptr to current VPB registry (one per VPB card)
	//VPBHANDLE *phandles;
	USHORT i,j;

	// validate arguments 
	assert(num != NULL);

	ci = new CONF_INFO;
	//verbose(1);

	config_data = vpb_read_confll(ci);
	if( config_data ){
		if (ci->verbosity == 1){
			verbose(1);
		}
		mprintf("We have a valid config file combination...\n");
		mprintf("\tNumber of OpenPri cards: %d\n",ci->OpenPri.cards);
		mprintf("\tNumber of OpenSwitch cards: %d\n",ci->OpenSwitch.cards);
		mprintf("\tNumber of OpenLine cards: %d\n",ci->OpenLine.cards);
		//mprintf("\tNumber of OpenPci cards: %d\n",ci->OpenPci.cards);
		mprintf("\tNumber of vtcore cards: %d\n",ci->vtcore.cards);
		//vpb_num_cards=ci->OpenPri.cards+ci->OpenSwitch.cards
		//             +ci->OpenLine.cards+ci->OpenPci.cards
		//             + ci->vtcore.cards;
		vpb_num_cards = ci->OpenPri.cards + ci->OpenSwitch.cards
			      + ci->OpenLine.cards + ci->vtcore.cards;
		mprintf("\tSum Total number of Cards: %d\n",vpb_num_cards);
		if (vpb_num_cards != ci->cardcount){
			mprintf("** Number of cards in vpb.conf[%d] doesnt match Sum Total[%d]! **\n",
                                ci->cardcount,vpb_num_cards);
		}
		// Set defaultsused as we are using the conf file
		defaultsused=0;
		// set vpb_model to -2 to indicate to use conf_info
		vpb_model=-2;
	}

	// if defaults have already been overwritten by call to vpb_config
	// ignore env vars
	if (defaultsused) {
		// attempt to autodetect hardware card type (can be overridden
		// by env variables
		#ifndef __FreeBSD__
		vpb_model =  GenericGetCardType();
		if (vpb_model != -1)
			mprintf("vpb_model = %d\n", vpb_model);
		else
			mprintf("Couldn't auto-detect board, trying environment..\n");
		#endif

		// check for environment variables that override defaults
		char *new_firmware, *new_model, *new_base, *new_bal1, *new_bal2;
		char *new_bal3, *new_RecordGain,*new_PlayGain;
		char *new_hostec, *new_dighyb, *new_verbose, *new_dtmf, *new_recordgaindown;

		// environment variable controlled message logging

		new_verbose = getenv("VPB_VERBOSE");
		if (new_verbose) {
			// if argument == 1, just switch on console logging
			if (strcmp(new_verbose, "1") == 0) {
				verbose(1);
			}
			else {
				// otherwise use file logging
				verbose_file(1, new_verbose);
			}
		}

		new_model = getenv("VPB_MODEL");
		if (new_model) {
			if (strcmp(new_model, "VPB4")==0) {
				vpb_model = VPB_VPB4;
				vpb_firmware_file = FIRMWARE_FILE_ISA;
			}
			if (strcmp(new_model, "VPB8L")==0) {
				vpb_model = VPB_VPB8L;
				vpb_firmware_file = FIRMWARE_FILE_ISA_VPB8L;
			}
			if (strcmp(new_model, "V4PCI")==0) {
				vpb_model = VPB_V4PCI;
				vpb_firmware_file = FIRMWARE_FILE_PCI;
			}
			if (strcmp(new_model, "V4LOG")==0) {
				vpb_model = VPB_V4LOG;
				vpb_firmware_file = FIRMWARE_FILE_PCI;
			}
			if (strcmp(new_model, "V12PCI")==0) {
				vpb_model = VPB_V12PCI;
				vpb_firmware_file = "";
			}
			if (strcmp(new_model, "V6PCI")==0) {
				vpb_model = VPB_V12PCI;
				//vpb_model = VPB_V6PCI;
				vpb_firmware_file = "";
			}
			if (strcmp(new_model, "PRI")==0) {
				vpb_model = VPB_PRI;
				vpb_firmware_file = "";
			}
		}
		if (vpb_model == -1){
			fprintf(stderr,
				"\nERROR No Voicetronix board or driver detected\n");
			exit(1);
		}

		new_firmware = getenv("VPB_FIRMWARE");
		if (new_firmware) {
			vpb_firmware_file = new_firmware;
		}

		new_base = getenv("VPB_BASE");
		if (new_base) {
			int base;
			sscanf(new_base, "%x", &base);
			def_base[0] = base;
		}

		// codec balance values specified by env vars

		hybdefs_overwrite = 0;
		new_bal1 = getenv("VPB_BAL1");
		if (new_bal1) {
			sscanf(new_bal1, "%x", &def_bal1);
			mprintf("new bal1 = 0x%x\n", def_bal1);
			hybdefs_overwrite = 1;
		}
		new_bal2 = getenv("VPB_BAL2");
		if (new_bal2) {
			sscanf(new_bal2, "%x", &def_bal2);
			mprintf("new bal2 = 0x%x\n", def_bal2);
			hybdefs_overwrite = 1;
		}
		new_bal3 = getenv("VPB_BAL3");
		if (new_bal3) {
			sscanf(new_bal3, "%x", &def_bal3);
			mprintf("new bal3 = 0x%x\n", def_bal3);
			hybdefs_overwrite = 1;
		}

		// optional HP filter on Tx and RX
		if (getenv("VPB_HP")) {
			def_hp = 1;
		}

		// Record and Play gains
		new_RecordGain = getenv("VPB_RG");
		if (new_RecordGain) {
			sscanf(new_RecordGain, "%f", &defRecordGain);
			n_rg=1;
			mprintf("new Record Gain = %f\n", defRecordGain);
		}
		new_PlayGain = getenv("VPB_PG");
		if (new_PlayGain) {
			sscanf(new_PlayGain, "%f", &defPlayGain);
			mprintf("new Play Gain = %f\n", defRecordGain);
			n_pg=1;
		}
		new_RecordGain = getenv("VPB_SRG");
		if (new_RecordGain) {
			sscanf(new_RecordGain, "%f", &defSRecordGain);
			n_rg=1;
		}
		new_PlayGain = getenv("VPB_SPG");
		if (new_PlayGain) {
			sscanf(new_PlayGain, "%f", &defSPlayGain);
			n_pg=1;
		}

		// host echo canceller
		new_hostec = getenv("VPB_HOSTEC");
		if (new_hostec) {
			def_hostec = DSPFIFO_ECHO;
			sscanf(new_hostec, "%d", &def_echo_en);
			mprintf("Host echo canc, enable = %d\n", def_echo_en);
		}

		// digital hybrid balance filter, note that we must enable hostec
		// so that dspfifo ECHO_THREADs run
		new_dighyb = getenv("VPB_DIGHYB");
		if (new_dighyb) {
			def_dighyb = 1;
			def_hostec = DSPFIFO_ECHO;
			def_dighybfile = new_dighyb;
			mprintf("digital hybrid balance filter enabled\n");
			mprintf("expected impulse responsefiles = %s0 %s1 ....\n",
			def_dighybfile);
		}

		// env variable for DTMF length in ms

		new_dtmf = getenv("VPB_DTMF");
		if (new_dtmf) {
			vpbdial_change_dtmf_length(atoi(new_dtmf), atoi(new_dtmf));
			mprintf("new dtmf on/off time = %s\n", new_dtmf);
		}

		// env DTMF cut through HW Record gain override
		new_recordgaindown = getenv("VPB_CUTTHROUGH");
		if (new_recordgaindown) {
			sscanf(new_recordgaindown, "%f", &defRecordGainDown);
			mprintf("Enabling Record Gain Down for Cut Through [%f]\n",defRecordGainDown);
		}

	} // end env variable scanning..............

	// auto-detect number of cards in system..........

	#ifdef __FreeBSD__
	*num = vpb_num_cards;
	#else
	if (defaultsused) {
		if ((vpb_model == VPB_V4PCI)||(vpb_model == VPB_V4LOG)
		    || (vpb_model == VPB_V12PCI)|| (vpb_model == VPB_V6PCI) ||(vpb_model == 69)) {
			// auto-determine number of cards if PCI
			int numCards;

			// I dont like this, locks to card model!!
			//numCards = GenericGetNumCards(vpb_model);
			#ifdef __linux__
			numCards = GenericGetNumCards(&cards);
			#else
			numCards = GenericGetNumCards(vpb_model);
			#endif
			*num = numCards;
		}
	}
	else {
		// hard code number of cards if defaults over-ridden
		*num = vpb_num_cards;
	}
	#endif

	if (*num==0)
	  throw Wobbly(VPBREG_NO_VPB_DEVICES_DETECTED);

	// Allocate memory for VPB DSP registry
	//reg = (VPBREG*)malloc(sizeof(VPBREG)*(*num));
	reg = (VPBREG*)calloc((*num),sizeof(VPBREG));
	assert(reg != NULL);

	// Allocate memory for VPB Handle map
	handles = (VPBHANDLE*)malloc(sizeof(VPBHANDLE)*(*num)*MAXCH);

	for (i=0;i<(*num)*MAXCH;i++){
		handles[i].board=-1;
		handles[i].channel=-1;
	}

	mprintf("number of cards [%d]\n", *num);
	for(i=0; i<*num; i++) {
		// set up VPB Reg info
		preg = &reg[i];
		preg->cardnum=i;
		if( config_data ) preg->useconf=1;

		// Check model
		if ((vpb_model > 0)&&(vpb_model<69) ){
			loc_model = vpb_model;
			preg->cardtypnum=i;
			mprintf("Card Number = %d(%d) Model = %d \n",
                                preg->cardnum, preg->cardtypnum, loc_model);
		}
		else if (cards != NULL){
			loc_model = cards[i];
			mprintf("Card Number = %d Model = %d (auto detect multi)\n",
                                i, loc_model);
		}
		else {
			// Determine model from config
			switch(ci->cards[i].type){
				case VPB_PRI:
					loc_model=VPB_PRI;
					preg->cardinfo = (void*)&ci->OpenPri.card[ci->cards[i].number];
					break;
				case VPB_V4PCI:
					loc_model=VPB_V4PCI;
					preg->cardinfo = (void*)&ci->OpenLine.card[ci->cards[i].number];
					break;
				case VPB_V4LOG:
					loc_model=VPB_V4LOG;
					preg->cardinfo = (void*)&ci->OpenLine.card[ci->cards[i].number];
					break;
				case VPB_V12PCI:
					if(ci->OpenSwitch.cards > 0){
						loc_model=VPB_V12PCI;
						preg->cardinfo = (void*)&ci->OpenSwitch.card[ci->cards[i].number];
					}
					else {
						loc_model=VPB_OSW;
						preg->cardinfo = (void*)&ci->vtcore.card[ci->cards[i].number];
					}
					break;
				case VPB_OPCI:
					loc_model=VPB_OPCI;
				    //XXX
				    #if 0
					if (ci->OpenPci.cards > 0)
						preg->cardinfo = (void*)&ci->OpenPci.card[ci->cards[i].number];
					else
				    #endif
						preg->cardinfo = (void*)&ci->vtcore.card[ci->cards[i].number];
					break;
				case VPB_OSW:
					loc_model=VPB_OSW;
					preg->cardinfo = (void*)&ci->vtcore.card[ci->cards[i].number];
					break;
			}
			mprintf("Card Number = %d Model = %d (from config file)\n", i, loc_model);
		}

		// initilize handle and bridge_intoarray
		for (j=0;j<MAXCH;j++){
			preg->handle[j]=-1;
			preg->listen_info[j].chan=-1;
			preg->listen_info[j].board=-1;
		}

		switch(loc_model) {
			case VPB_VPB4:
				preg->numch = 4;
				preg->model = VPB_VPB4;
				preg->base = (USHORT)vpb_base[i];
				preg->szmess = SIZE_MESS_Q_OL4;
				break;
			case VPB_V4PCI:
				preg->numch = 4;
				preg->model = VPB_V4PCI;
				preg->base = i;
				if( config_data ){
					preg->cardtypnum = ci->cards[i].number;
				} else {
					preg->cardtypnum = preg->cardnum;
				}
				preg->szmess = SIZE_MESS_Q_OL4;
				break;
			case VPB_V4LOG:
				preg->numch = 4;
				preg->model = VPB_V4LOG;
				preg->base = i;
				if( config_data ){
					preg->cardtypnum = ci->cards[i].number;
				} else {
					preg->cardtypnum = preg->cardnum;
				}
				preg->szmess = SIZE_MESS_Q_OL4;
				break;
			case VPB_VPB8L:
				preg->numch = 8;
				preg->model = VPB_VPB8L;
				preg->base = (USHORT)vpb_base[i];
				preg->szmess = SIZE_MESS_Q_OL4;
				break;
			case VPB_V12PCI:
				preg->numch = 12;
				preg->model = VPB_V12PCI;
				preg->base = i;
				if( config_data ){
					preg->cardtypnum = ci->cards[i].number;
				} else {
					preg->cardtypnum = preg->cardnum;
				}
				preg->szmess = SIZE_MESS_Q_HDSP;
				break;
			case VPB_V6PCI:
				preg->numch = 12;	//first 6 ports empty
				preg->model = VPB_V12PCI;
				preg->base = i;
				if( config_data ){
					preg->cardtypnum = ci->cards[i].number;
				} else {
					preg->cardtypnum = preg->cardnum;
				}
				preg->szmess = SIZE_MESS_Q_HDSP;
				// OS_BASE=6;
				break;
			case VPB_PRI:
				preg->model = VPB_PRI;
				preg->numch = (ci->OpenPri.card[ci->cards[i].number].if_type==VPB_PRI_E1?30:23);
				preg->base = i;
				preg->cardtypnum = ci->cards[i].number;
				preg->szmess = SIZE_MESS_Q_HDSP;
				break;
			case VPB_OSW:
			case VPB_OPCI:
				preg->model = loc_model;
				preg->numch = ci->vtcore.card[ci->cards[i].number].channels;
				preg->base = i;
				preg->cardtypnum = ci->cards[i].number;
				preg->szmess = SIZE_MESS_Q_HDSP;
				break;
			default:
				assert(0);
				break;
		}
		for(j=0; j<MAXCH; j++) {
			preg->szrxdf[j] = SIZE_RX_Q;
			preg->sztxdf[j] = SIZE_RX_Q;
		}
		preg->lsf = LENGTH_FRAME;

		switch(loc_model) {
			case VPB_VPB4:
				preg->ddmodel = DD_ISA_PORT;
				preg->defRecordGain = DEF_RECORD_GAIN_ISA;
				preg->defPlayGain = DEF_PLAY_GAIN_ISA;
				break;
			case VPB_V4PCI:
				preg->ddmodel = DD_PCI;
				preg->defRecordGain = defRecordGain;
				preg->defPlayGain = defPlayGain;
				break;
			case VPB_V4LOG:
				preg->ddmodel = DD_PCI;
				preg->defRecordGain = defRecordGain;
				preg->defPlayGain = defPlayGain;
				break;
			case VPB_VPB8L:
				preg->ddmodel = DD_ISA_PORT;
				break;
			case VPB_V12PCI:
				preg->ddmodel = DD_V12PCI;
				if (n_rg ==1)   //if we have one from the command line 
					preg->defRecordGain = defRecordGain;
				else
					preg->defRecordGain = DEF_RECORD_GAIN_V12PCI;
				if (n_pg ==1)   //if we have one from the command line
					preg->defPlayGain = defPlayGain;
				else
					preg->defPlayGain = DEF_PLAY_GAIN_V12PCI;
				preg->defSRecordGain = defSRecordGain;
				preg->defSPlayGain = defSPlayGain;
				break;
			case VPB_V6PCI:
				preg->ddmodel = DD_V12PCI;
				if (n_rg ==1)   //if we have one from the command line 
					preg->defRecordGain = defRecordGain;
				else
					preg->defRecordGain = DEF_RECORD_GAIN_V12PCI;
				if (n_pg ==1)	//if we have one from the command line
					preg->defPlayGain = defPlayGain;
				else
					preg->defPlayGain = DEF_PLAY_GAIN_V12PCI;
				preg->defSRecordGain = defSRecordGain;
				preg->defSPlayGain = defSPlayGain;
				break;
			case VPB_PRI:
				preg->ddmodel = DD_PRI;
				if (n_rg ==1)   //if we have one from the command line 
					preg->defRecordGain = defRecordGain;
				else
					preg->defRecordGain = DEF_RECORD_GAIN_PRI;
				if (n_pg ==1)	//if we have one from the command line
					preg->defPlayGain = defPlayGain;
				else
					preg->defPlayGain = DEF_PLAY_GAIN_PRI;
				break;
			case VPB_OPCI:
				preg->ddmodel = DD_VTCORE;
				preg->defRecordGain = 0;
				preg->defPlayGain = 0;
				break;
			case VPB_OSW:
				preg->ddmodel = DD_VTCORE;
				preg->defRecordGain = DEF_RECORD_GAIN_V12PCI;
				preg->defPlayGain = DEF_PLAY_GAIN_V12PCI;
				preg->defSRecordGain = DEF_RECORD_GAIN_SV12PCI;
				preg->defSPlayGain = DEF_PLAY_GAIN_SV12PCI;
				break;
			default:
				assert(0);
				break;
		}

		preg->defRecordGainDown = defRecordGainDown;

		mprintf("defRecordGain=%f\n",preg->defRecordGain);
		mprintf("defPlayGain=%f\n",preg->defPlayGain);
		preg->szRelayBuf = 0; // unused

		preg->hybdefs_overwrite = hybdefs_overwrite;
		preg->defbal1 = def_bal1;
		preg->defbal2 = def_bal2;
		preg->defbal3 = def_bal3;
		preg->hp_enable = def_hp;
		preg->echo_mode = def_hostec;
		preg->echo_en = def_echo_en;
		preg->dighyb = def_dighyb;
		preg->dighybfile = def_dighybfile;

		// init ptrs to PC copy of FIFO state variables
		preg->dnmess = NULL;
		preg->upmess = NULL;
		for(j=0; j<MAXCH; j++) {
			preg->rxdf[j] = NULL;
			preg->txdf[j] = NULL;
		}
		// set up firmware path and file name
		strcpy(preg->firm, vpb_firmware_file);

		// Initialize rest of structures
		preg->data = 0;
		for(j=0; j<MAXCH; j++) {
			preg->chstat[j] = 0;
			preg->a_rxdf[j] = 0;
			preg->a_txdf[j] = 0;
			preg->ec[j] = NULL;
		}

		preg->a_dnmess = 0;
		preg->a_upmess = 0;

		for(j=0; j<12; j++)
			preg->mdate[j] = '\0';
		for(j=0; j<8; j++)
			preg->revision[j] = '\0';
		for(j=0; j<10; j++)
			preg->serial_n[j] = '\0';
	}

	if (cards != NULL)
		free(cards);
}

int vpbreg_get_model() {
	return vpb_model;
}
int vpbreg_get_model(int h) {
	USHORT b,ch;
	int model=-2;
	maphndletodev(h, &b, &ch);
	VPBREG *vr = vpb_c->vpbreg(b);
	model = vr->model;
	//mprintf("vpbreg_get_model: Channel [%d] on Board [%d] is model [%d]\n",ch,b,model);
	return model;
}

VPBREG *vpbreg_get_reg(int b)
{
	return vpb_c->vpbreg(b);
}

void vpbreg_load_default_tones(int /* numch */)
{
	// defaults for tone detectors should be loaded here
}

